#!/usr/bin/env python3

"""
test.unit_tests_d.ut_testdep: test dependency unit tests for the MMGen suite
"""

import sys,os
from subprocess import run,DEVNULL

from mmgen.util import ymsg,bmsg
from ..include.common import cfg

sec = 'deadbeef' * 8

class unit_tests:

	altcoin_deps = ('pycoin','monero_python','keyconv','zcash_mini','ethkey','ssh_socks_proxy')
	win_skip = ('losetup','zcash_mini')

	def pylint(self,name,ut):
		try:
			return run(['pylint','--version'],stdout=None if cfg.verbose else DEVNULL).returncode == 0
		except OSError as e:
			ymsg('  ' + str(e))
			bmsg("  Install pylint with 'python3 -m pip install pylint'")
			return False

	def core_repo(self,name,ut):
		crr = os.getenv('CORE_REPO_ROOT')
		if not crr or not os.path.exists(os.path.join(crr,'src/test/data/tx_valid.json')):
			ymsg('CORE_REPO_ROOT not set, or does not point to Bitcoin Core repository')
			return False
		return True

	def losetup(self,name,ut):
		os.stat('/dev/loop0')
		run(['/sbin/losetup','-f'],check=True,stdout=DEVNULL)
		return True

	def pycoin(self,name,ut):
		from pycoin.networks.registry import network_for_netcode as nfnc
		network = nfnc('btc')
		key = network.keys.private(secret_exponent=int(sec,16),is_compressed=True)
		hash160_c = key.hash160(is_compressed=True)
		network.address.for_p2pkh_wit(hash160_c)
		return True

	def monero_python(self,name,ut):
		from mmgen.util2 import load_cryptodomex
		load_cryptodomex()
		from monero.seed import Seed
		Seed('deadbeef' * 8).public_address()
		return True

	def keyconv(self,name,ut):
		run(['keyconv','-G','ltc'],stdout=DEVNULL,stderr=DEVNULL,check=True)
		return True

	def zcash_mini(self,name,ut):
		run(['zcash-mini'],stdout=DEVNULL,check=True)
		return True

	def ethkey(self,name,ut):
		if sys.platform == 'linux' and os.uname().machine != 'x86_64':
			distro = [l for l in open('/etc/os-release').read().split('\n') if l.startswith('ID=')][0][3:]
			if distro != 'archarm':
				ut.skip_msg(f'distro {distro!r} on architecture {os.uname().machine!r}')
				return True
		from test.include.common import get_ethkey
		get_ethkey()
		return True

	def ssh_socks_proxy(self,name,ut):
		from test.cmdtest_py_d.ct_xmrwallet import CmdTestXMRWallet
		return CmdTestXMRWallet.init_proxy(external_call=True)



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, a command-line cryptocurrency wallet
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tx.status: transaction status class
"""

class Status:

	def __init__(self,parent_tx):
		self.tx = parent_tx



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
color: color handling for the MMGen suite
"""

_colors = {
	'black':       (  232,      (30,0) ),
	'red':         (  210,      (31,1) ),
	'green':       (  121,      (32,1) ),
	'yellow':      (  229,      (33,1) ),
	'blue':        (  75,       (34,1) ),
	'magenta':     (  205,      (35,1) ),
	'cyan':        (  122,      (36,1) ),

	'gray':        (  246,      (30,1) ),
	'orange':      (  216,      (31,1) ),
	'purple':      (  141,      (35,1) ),
	'pink':        (  218,      (35,1) ),

	'melon':       (  222,      (33,1) ),
	'brown':       (  173,      (33,0) ),
	'grndim':      (  108,      (32,0) ),

	'redbg':       ( (232,210), (30,101) ),
	'grnbg':       ( (232,121), (30,102) ),
	'yelbg':       ( (232,229), (30,103) ),
	'blubg':       ( (232,75),  (30,104) ),
}

def nocolor(s):
	return s

def set_vt100():
	'hack to put term into VT100 mode under MSWin'
	import sys
	if sys.platform == 'win32':
		from subprocess import run
		run([],shell=True)

def get_terminfo_colors(term=None):
	from subprocess import run,PIPE
	cmd = ['infocmp','-0']
	if term:
		cmd.append(term)

	try:
		cmdout = run(cmd,stdout=PIPE,check=True).stdout.decode()
	except:
		set_vt100()
		return None
	else:
		set_vt100()
		s = [e.split('#')[1] for e in cmdout.split(',') if e.startswith('colors')][0]
		from .util import is_hex_str
		if s.isdecimal():
			return int(s)
		elif s.startswith('0x') and is_hex_str(s[2:]):
			return int(s[2:],16)
		else:
			return None

def init_color(num_colors='auto'):
	assert num_colors in ('auto',8,16,256,0)

	import sys
	self = sys.modules[__name__]

	if num_colors == 'auto':
		import os
		if sys.platform == 'win32':
			# Force 256-color for MSYS2: terminal supports it, however infocmp reports 8-color.
			# We also avoid spawning a subprocess, leading to a subsequent OSError 22 when testing
			# with pexpect spawn.
			num_colors = 256
		elif (t := os.getenv('TERM')) and t.endswith('256color'):
			num_colors = 256
		else:
			num_colors = get_terminfo_colors() or 16

	reset = '\033[0m'
	if num_colors == 0:
		ncc = (lambda s: s).__code__
		for c in _colors:
			getattr(self,c).__code__ = ncc
	elif num_colors == 256:
		for c,e in _colors.items():
			start = (
				'\033[38;5;{};1m'.format(e[0]) if type(e[0]) == int else
				'\033[38;5;{};48;5;{};1m'.format(*e[0]) )
			getattr(self,c).__code__ = eval(f'(lambda s: "{start}" + s + "{reset}").__code__')
	elif num_colors in (8,16):
		for c,e in _colors.items():
			start = (
				'\033[{}m'.format(e[1][0]) if e[1][1] == 0 else
				'\033[{};{}m'.format(*e[1]) )
			getattr(self,c).__code__ = eval(f'(lambda s: "{start}" + s + "{reset}").__code__')

	set_vt100()

# Each color name must be bound to an independent stub function with its own
# address in memory.  The names themselves must never be redefined, since other
# modules could import them before init_color() is run.  Instead, the code
# objects of their associated functions are manipulated by init_color() to
# enable/disable color.

black   = lambda s: s
red     = lambda s: s
green   = lambda s: s
yellow  = lambda s: s
blue    = lambda s: s
magenta = lambda s: s
cyan    = lambda s: s

gray    = lambda s: s
orange  = lambda s: s
purple  = lambda s: s
pink    = lambda s: s

melon   = lambda s: s
brown   = lambda s: s
grndim  = lambda s: s

redbg   = lambda s: s
grnbg   = lambda s: s
yelbg   = lambda s: s
blubg   = lambda s: s



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>

"""
test.objtest_py_d.ot_ltc_mainnet: LTC mainnet test vectors for MMGen data objects
"""

from decimal import Decimal

from mmgen.amt import LTCAmt
from mmgen.addr import CoinAddr
from mmgen.key import WifKey,PrivKey
from mmgen.protocol import init_proto

from .ot_common import r16,r32
from ..include.common import cfg

proto = init_proto( cfg, 'ltc', need_amt=True )

tests = {
	'LTCAmt': {
		'bad':  ('-3.2','0.123456789',123,'123L','88000000',80999999.12345678),
		'good': (('80999999.12345678',Decimal('80999999.12345678')),)
	},
	'CoinAddr': {
		'bad':  (
			{'addr':1,   'proto':proto},
			{'addr':'x', 'proto':proto},
			{'addr':'я', 'proto':proto},
		),
		'good':  (
			{'addr':'LXYx4j8PDGE8GEwDFnEQhcLyHFGsRxSJwt', 'proto':proto},
			{'addr':'MEnuCzUGHaQx9fK5WYvLwR1NK4SAo8HmSr', 'proto':proto},
		),
	},
	'WifKey': {
		'bad': (
			{'proto':proto, 'wif':1},
			{'proto':proto, 'wif':[]},
			{'proto':proto, 'wif':'\0'},
			{'proto':proto, 'wif':'\1'},
			{'proto':proto, 'wif':'я'},
			{'proto':proto, 'wif':'g'},
			{'proto':proto, 'wif':'gg'},
			{'proto':proto, 'wif':'FF'},
			{'proto':proto, 'wif':'f00'},
			{'proto':proto, 'wif':r16.hex()},
			{'proto':proto, 'wif':'2MspvWFjBbkv2wzQGqhxJUYPCk3Y2jMaxLN'},
		),
		'good': (
			{'proto':proto, 'wif':'6udBAGS6B9RfGyvEQDkVDsWy3Kqv9eTULqtEfVkJtTJyHdLvojw',  'ret_idx':1},
			{'proto':proto, 'wif':'T7kCSp5E71jzV2zEJW4q5qU1SMB5CSz8D9VByxMBkamv1uM3Jjca', 'ret_idx':1},
		)
	},
	'PrivKey': {
		'bad': (
			{'proto':proto, 'wif':1},
			{'proto':proto, 'wif':'1'},
			{'proto':proto, 'wif':'cMsqcmDYZP1LdKgqRh9L4ZRU9br28yvdmTPwW2YQwVSN9aQiMAoR'},
			{'proto':proto, 's':r32,'wif':'6ufJhtQQiRYA3w2QvDuXNXuLgPFp15i3HR1Wp8An2mx1JnhhJAh'},
			{'proto':proto, 'pubkey_type':'std'},
			{'proto':proto, 's':r32},
			{'proto':proto, 's':r32,'compressed':'yes'},
			{'proto':proto, 's':r32,'compressed':'yes','pubkey_type':'std'},
			{'proto':proto, 's':r32,'compressed':True,'pubkey_type':'nonstd'},
			{'proto':proto, 's':r32+b'x','compressed':True,'pubkey_type':'std'}
		),
		'good': (
			{'proto':proto, 'wif':'6ufJhtQQiRYA3w2QvDuXNXuLgPFp15i3HR1Wp8An2mx1JnhhJAh',
			'ret':bytes.fromhex('470a974ffca9fca1299b706b09142077bea3acbab6d6480b87dbba79d5fd279b')},
			{'proto':proto, 'wif':'T41Fm7J3mtZLKYPMCLVSFARz4QF8nvSDhLAfW97Ds56Zm9hRJgn8',
			'ret':bytes.fromhex('1c6feab55a4c3b4ad1823d4ecacd1565c64228c01828cf44fb4db1e2d82c3d56')},
			{'proto':proto, 's':r32,'compressed':False,'pubkey_type':'std','ret':r32},
			{'proto':proto, 's':r32,'compressed':True,'pubkey_type':'std','ret':r32}
		)
	},
}



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, a command-line cryptocurrency wallet
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.eth.tx.base: Ethereum base transaction class
"""

from collections import namedtuple

from ....tx import base as TxBase
from ....obj import HexStr,Int

class Base(TxBase.Base):

	rel_fee_desc = 'gas price'
	rel_fee_disp = 'gas price in Gwei'
	txobj  = None # ""
	dfl_gas = 21000               # an approximate number, used for fee estimation purposes
	dfl_start_gas = 21000         # the actual startgas amt used in the transaction
	                              # for simple sends with no data, gas = start_gas = 21000
	contract_desc = 'contract'
	usr_contract_data = HexStr('')
	disable_fee_check = False

	# given absolute fee in ETH, return gas price in Gwei using self.gas
	def fee_abs2rel(self,abs_fee,to_unit='Gwei'):
		ret = self.proto.coin_amt(int(abs_fee.toWei() // self.gas.toWei()),'wei')
		self.cfg._util.dmsg(f'fee_abs2rel() ==> {ret} ETH')
		return ret if to_unit == 'eth' else ret.to_unit(to_unit,show_decimal=True)

	# given rel fee (gasPrice) in wei, return absolute fee using self.gas (Ethereum-only method)
	def fee_gasPrice2abs(self,rel_fee):
		assert isinstance(rel_fee,int), f'{rel_fee!r}: incorrect type for fee estimate (not an integer)'
		return self.proto.coin_amt(rel_fee * self.gas.toWei(),'wei')

	def is_replaceable(self):
		return True

	async def get_receipt(self,txid):
		rx = await self.rpc.call('eth_getTransactionReceipt','0x'+txid) # -> null if pending
		if not rx:
			return None
		tx = await self.rpc.call('eth_getTransactionByHash','0x'+txid)
		return namedtuple('exec_status',['status','gas_sent','gas_used','gas_price','contract_addr','tx','rx'])(
			status        = Int(rx['status'],16), # zero is failure, non-zero success
			gas_sent      = Int(tx['gas'],16),
			gas_used      = Int(rx['gasUsed'],16),
			gas_price     = self.proto.coin_amt(int(tx['gasPrice'],16),from_unit='wei'),
			contract_addr = self.proto.coin_addr(rx['contractAddress'][2:]) if rx['contractAddress'] else None,
			tx            = tx,
			rx            = rx,
		)

	def check_serialized_integrity(self): # TODO
		return True

class TokenBase(Base):
	dfl_gas = 52000
	dfl_start_gas = 60000
	contract_desc = 'token contract'



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
mmgen-txcreate: Create a cryptocoin transaction with MMGen- and/or non-MMGen
                inputs and outputs
"""

from .cfg import gc,Config
from .util import fmt_list,async_run

opts_data = {
	'sets': [('yes', True, 'quiet', True)],
	'text': {
		'desc': f'Create a transaction with outputs to specified coin or {gc.proj_name} addresses',
		'usage':   '[opts]  [<addr,amt> ...] <change addr, addrlist ID or addr type> [addr file ...]',
		'options': """
-h, --help            Print this help message
--, --longhelp        Print help message for long options (common options)
-a, --autosign        Create a transaction for offline autosigning (see
                      ‘mmgen-autosign’). The removable device is mounted and
                      unmounted automatically
-A, --fee-adjust=  f  Adjust transaction fee by factor 'f' (see below)
-B, --no-blank        Don't blank screen before displaying unspent outputs
-c, --comment-file=f  Source the transaction's comment from file 'f'
-C, --fee-estimate-confs=c Desired number of confirmations for fee estimation
                      (default: {cfg.fee_estimate_confs})
-d, --outdir=      d  Specify an alternate directory 'd' for output
-D, --contract-data=D Path to hex-encoded contract data (ETH only)
-E, --fee-estimate-mode=M Specify the network fee estimate mode.  Choices:
                      {fe_all}.  Default: {fe_dfl!r}
-f, --fee=         f  Transaction fee, as a decimal {cu} amount or as
                      {fu} (an integer followed by {fl!r}).
                      See FEE SPECIFICATION below.  If omitted, fee will be
                      calculated using network fee estimation.
-g, --gas=         g  Specify start gas amount in Wei (ETH only)
-i, --info            Display unspent outputs and exit
-I, --inputs=      i  Specify transaction inputs (comma-separated list of
                      MMGen IDs or coin addresses).  Note that ALL unspent
                      outputs associated with each address will be included.
-l, --locktime=    t  Lock time (block height or unix seconds) (default: 0)
-L, --autochg-ignore-labels Ignore labels when autoselecting change addresses
-m, --minconf=     n  Minimum number of confirmations required to spend
                      outputs (default: 1)
-q, --quiet           Suppress warnings; overwrite files without prompting
-R, --no-rbf          Make transaction non-replaceable (non-replace-by-fee
                      according to BIP 125)
-v, --verbose         Produce more verbose output
-V, --vsize-adj=   f  Adjust transaction's estimated vsize by factor 'f'
-y, --yes             Answer 'yes' to prompts, suppress non-essential output
-X, --cached-balances Use cached balances (Ethereum only)
""",
		'notes': '\n{c}\n{F}\n{x}',
	},
	'code': {
		'options': lambda cfg,proto,help_notes,s: s.format(
			fu=help_notes('rel_fee_desc'),
			fl=help_notes('fee_spec_letters'),
			fe_all=fmt_list(cfg._autoset_opts['fee_estimate_mode'].choices,fmt='no_spc'),
			fe_dfl=cfg._autoset_opts['fee_estimate_mode'].choices[0],
			cu=proto.coin,
			cfg=cfg),
		'notes': lambda cfg,help_notes,s: s.format(
			c = help_notes('txcreate'),
			F = help_notes('fee'),
			x = help_notes('txcreate_examples') )
	}
}

cfg = Config(opts_data=opts_data)

async def main():

	if cfg.autosign:
		from .tx.util import init_removable_device
		from .autosign import Signable
		asi = init_removable_device(cfg)
		asi.do_mount()
		Signable.automount_transaction(asi).check_create_ok()

	from .tx import NewTX
	tx1 = await NewTX(cfg=cfg,proto=cfg._proto)

	from .rpc import rpc_init
	tx1.rpc = await rpc_init(cfg)

	tx2 = await tx1.create(
		cmd_args = cfg._args,
		locktime = int(cfg.locktime or 0),
		do_info  = cfg.info )

	tx2.file.write(
		outdir                = asi.txauto_dir if cfg.autosign else None,
		ask_write             = not cfg.yes,
		ask_overwrite         = not cfg.yes,
		ask_write_default_yes = False)

async_run(main())



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, a command-line cryptocurrency wallet
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
altcointest.py - Test constants for Bitcoin-derived altcoins
"""

import sys

try:
	from include import test_init
except ImportError:
	from test.include import test_init

from mmgen.cfg import gc,Config
from mmgen.util import msg
from mmgen.altcoin.params import CoinInfo

def test_equal(desc,a,b,*cdata):
	if type(a) is int:
		a = hex(a)
		b = hex(b)
	(network,coin,_,b_desc,verbose) = cdata
	if verbose:
		msg(f'  {desc:20}: {a!r}')
	if a != b:
		raise ValueError(
			f'{desc.capitalize()}s for {coin.upper()} {network} do not match:\n  CoinInfo: {a}\n  {b_desc}: {b}' )

class TestCoinInfo(CoinInfo):

	# Sources (see CoinInfo) that are in agreement for these coins
	# No check for segwit, p2sh check skipped if source doesn't support it
	cross_checks = {
		'2GIVE':  ['wn'],
		'42':     ['vg','wn'],
		'611':    ['wn'],
		'AC':     ['lb','vg'],
		'ACOIN':  ['wn'],
		'ALF':    ['wn'],
		'ANC':    ['vg','wn'],
		'APEX':   ['wn'],
		'ARCO':   ['wn'],
		'ARG':    ['pc'],
		'AUR':    ['vg','wn'],
		'BCH':    ['wn'],
		'BLK':    ['lb','vg','wn'],
		'BQC':    ['vg','wn'],
		'BSTY':   ['wn'],
		'BTC':    ['lb','vg','wn'],
		'BTCD':   ['lb','vg','wn'],
		'BUCKS':  ['wn'],
		'CASH':   ['wn'],
		'CBX':    ['wn'],
		'CCN':    ['lb','vg','wn'],
		'CDN':    ['lb','vg','wn'],
		'CHC':    ['wn'],
		'CLAM':   ['lb','vg'],
		'CON':    ['vg','wn'],
		'CPC':    ['wn'],
		'DASH':   ['lb','pc','vg','wn'],
		'DCR':    ['pc'],
		'DFC':    ['pc'],
		'DGB':    ['lb','vg'],
		'DGC':    ['lb','vg','wn'],
		'DOGE':   ['lb','pc','vg','wn'],
		'DOGED':  ['lb','vg','wn'],
		'DOPE':   ['lb','vg'],
		'DVC':    ['vg','wn'],
		'EFL':    ['lb','vg','wn'],
		'EMC':    ['vg'],
		'EMD':    ['wn'],
		'ESP':    ['wn'],
		'FAI':    ['pc'],
		'FC2':    ['wn'],
		'FIBRE':  ['wn'],
		'FJC':    ['wn'],
		'FLO':    ['wn'],
		'FLT':    ['wn'],
		'FST':    ['wn'],
		'FTC':    ['lb','pc','vg','wn'],
		'GCR':    ['lb','vg'],
		'GOOD':   ['wn'],
		'GRC':    ['vg','wn'],
		'GUN':    ['vg','wn'],
		'HAM':    ['vg','wn'],
		'HTML5':  ['wn'],
		'HYP':    ['wn'],
		'ICASH':  ['wn'],
		'INFX':   ['wn'],
		'IPC':    ['wn'],
		'JBS':    ['lb','pc','vg','wn'],
		'JUDGE':  ['wn'],
		'LANA':   ['wn'],
		'LAT':    ['wn'],
		'LDOGE':  ['wn'],
		'LMC':    ['wn'],
		'LTC':    ['lb','vg','wn'],
		'MARS':   ['wn'],
		'MEC':    ['pc','wn'],
		'MINT':   ['wn'],
		'MOBI':   ['wn'],
		'MONA':   ['lb','vg'],
		'MOON':   ['wn'],
		'MUE':    ['lb','vg'],
		'MXT':    ['wn'],
		'MYR':    ['pc'],
		'MYRIAD': ['vg','wn'],
		'MZC':    ['lb','pc','vg','wn'],
		'NEOS':   ['lb','vg'],
		'NEVA':   ['wn'],
		'NKA':    ['wn'],
		'NLG':    ['vg','wn'],
		'NMC':    ['lb','vg'],
		'NVC':    ['lb','vg','wn'],
		'OK':     ['lb','vg'],
		'OMC':    ['vg','wn'],
		'ONION':  ['vg','wn'],
		'PART':   ['wn'],
		'PINK':   ['vg','wn'],
		'PIVX':   ['wn'],
		'PKB':    ['lb','vg','wn'],
		'PND':    ['lb','vg','wn'],
		'POT':    ['lb','vg','wn'],
		'PPC':    ['lb','vg','wn'],
		'PTC':    ['vg','wn'],
		'PXC':    ['wn'],
		'QRK':    ['wn'],
		'RAIN':   ['wn'],
		'RBT':    ['wn'],
		'RBY':    ['lb','vg'],
		'RDD':    ['vg','wn'],
		'RIC':    ['pc','vg','wn'],
		'SDC':    ['lb','vg'],
		'SIB':    ['wn'],
		'SMLY':   ['wn'],
		'SONG':   ['wn'],
		'SPR':    ['vg','wn'],
		'START':  ['lb','vg'],
		'SYS':    ['wn'],
		'TAJ':    ['wn'],
		'TIT':    ['wn'],
		'TPC':    ['lb','vg'],
		'TRC':    ['wn'],
		'TTC':    ['wn'],
		'TX':     ['wn'],
		'UNO':    ['pc','vg','wn'],
		'VIA':    ['lb','pc','vg','wn'],
		'VPN':    ['lb','vg'],
		'VTC':    ['lb','vg','wn'],
		'WDC':    ['vg','wn'],
		'WISC':   ['wn'],
		'WKC':    ['vg','wn'],
		'WSX':    ['wn'],
		'XCN':    ['wn'],
		'XGB':    ['wn'],
		'XPM':    ['lb','vg','wn'],
		'XST':    ['wn'],
		'XVC':    ['wn'],
		'ZET':    ['wn'],
		'ZOOM':   ['lb','vg'],
		'ZRC':    ['lb','vg']
	}

	@classmethod
	def verify_leading_symbols(cls,quiet=False,verbose=False):

		for network in ('mainnet','testnet'):
			for coin in [e.symbol for e in cls.coin_constants[network]]:
				e = cls.get_entry(coin,network)
				cdata = (network,coin,e,'Computed value',verbose)

				if not quiet:
					msg(f'{coin} {network}')

				vn_info = e.p2pkh_info
				ret = cls.find_addr_leading_symbol(vn_info[0])
				test_equal('P2PKH leading symbol',vn_info[1],ret,*cdata)

				vn_info = e.p2sh_info
				if vn_info:
					ret = cls.find_addr_leading_symbol(vn_info[0])
					test_equal('P2SH leading symbol',vn_info[1],ret,*cdata)

	@classmethod
	def verify_core_coin_data(cls,cfg,quiet=False,verbose=False):
		from mmgen.protocol import CoinProtocol,init_proto

		for network in ('mainnet','testnet'):
			for coin in gc.core_coins:
				e = cls.get_entry(coin,network)
				if e:
					proto = init_proto( cfg, coin, network=network )
					cdata = (network,coin,e,type(proto).__name__,verbose)
					if not quiet:
						msg(f'Verifying {coin.upper()} {network}')

					if coin != 'bch': # TODO
						test_equal('coin name',e.name,proto.name,*cdata)

					if e.trust_level != -1:
						test_equal('Trust level',e.trust_level,CoinProtocol.coins[coin].trust_level,*cdata)

					test_equal(
						'WIF version number',
						e.wif_ver_num,
						int.from_bytes(proto.wif_ver_bytes['std'],'big'),
						*cdata )

					test_equal(
						'P2PKH version number',
						e.p2pkh_info[0],
						int.from_bytes(proto.addr_fmt_to_ver_bytes['p2pkh'],'big'),
						*cdata )

					test_equal(
						'P2SH version number',
						e.p2sh_info[0],
						int.from_bytes(proto.addr_fmt_to_ver_bytes['p2sh'],'big'),
						*cdata )

	# Data is one of the coin_constants lists above.  Normalize ints to hex of correct width, add
	# missing leading letters, set trust level from external_tests.
	# Insert a coin entry from outside source, set version info leading letters to '?' and trust level
	# to 0, then run TestCoinInfo.fix_table(data).  'has_segwit' field is updated manually for now.
	@classmethod
	def fix_table(cls,data):
		import re

		def myhex(n):
			return '0x{:0{}x}'.format(n,2 if n < 256 else 4)

		def fix_ver_info(e,k):
			e[k] = list(e[k])
			e[k][0] = myhex(e[k][0])
			s1 = cls.find_addr_leading_symbol(int(e[k][0][2:],16))
			m = f'Fixing leading address letter for coin {e["symbol"]} ({e[k][1]!r} --> {s1})'
			if e[k][1] != '?':
				assert s1 == e[k][1], f'First letters do not match! {m}'
			else:
				msg(m)
				e[k][1] = s1
			e[k] = tuple(e[k])

		old_sym = None
		for sym in sorted([e.symbol for e in data]):
			if sym == old_sym:
				msg(f'{sym!r}: duplicate coin symbol in data!')
				sys.exit(2)
			old_sym = sym

		tt = cls.create_trust_table()

		name_w = max(len(e.name) for e in data)
		fs = '\t({:%s} {:10} {:7} {:17} {:17} {:6} {}),' % (name_w+3)
		for e in data:
			e = e._asdict()
			e['wif_ver_num'] = myhex(e['wif_ver_num'])
			sym,trust = e['symbol'],e['trust_level']

			fix_ver_info(e,'p2pkh_info')
			if isinstance(e['p2sh_info'],tuple):
				fix_ver_info(e,'p2sh_info')

			for k in e.keys():
				e[k] = repr(e[k])
				e[k] = re.sub(r"'0x(..)'",r'0x\1',e[k])
				e[k] = re.sub(r"'0x(....)'",r'0x\1',e[k])
				e[k] = re.sub(r' ',r'',e[k]) + ('',',')[k != 'trust_level']

			if trust != -1:
				if sym in tt:
					src = tt[sym]
					if src != trust:
						msg(f'Updating trust for coin {sym!r}: {trust} -> {src}')
						e['trust_level'] = src
				else:
					if trust != 0:
						msg(f'Downgrading trust for coin {sym!r}: {trust} -> 0')
						e['trust_level'] = 0

				if sym in cls.cross_checks:
					if int(e['trust_level']) == 0 and len(cls.cross_checks[sym]) > 1:
						msg(f'Upgrading trust for coin {sym!r}: {e["trust_level"]} -> 1')
						e['trust_level'] = 1

			print(fs.format(*e.values()))
		msg(f'Processed {len(data)} entries')

	@classmethod
	def find_addr_leading_symbol(cls,ver_num,verbose=False):

		if ver_num == 0:
			return '1'

		def phash2addr(ver_num,pk_hash):
			from mmgen.proto.btc.common import b58chk_encode
			bl = ver_num.bit_length()
			ver_bytes = int.to_bytes(ver_num,bl//8 + bool(bl%8),'big')
			return b58chk_encode(ver_bytes + pk_hash)

		low = phash2addr(ver_num,b'\x00'*20)
		high = phash2addr(ver_num,b'\xff'*20)

		if verbose:
			print('low address:  ' + low)
			print('high address: ' + high)

		l1,h1 = low[0],high[0]
		return (l1,h1) if l1 != h1 else l1

	@classmethod
	def print_symbols(cls,include_names=False,reverse=False):
		for e in cls.coin_constants['mainnet']:
			if reverse:
				print(f'{e.symbol:6} {e.name}')
			else:
				name_w = max(len(e.name) for e in cls.coin_constants['mainnet'])
				print((f'{e.name:{name_w}} ' if include_names else '') + e.symbol)

	@classmethod
	def create_trust_table(cls):
		tt = {}
		mn = cls.external_tests['mainnet']
		for ext_prog in mn:
			assert len(set(mn[ext_prog])) == len(mn[ext_prog]), f'Duplicate entry in {ext_prog!r}!'
			for coin in mn[ext_prog]:
				if coin in tt:
					tt[coin] += 1
				else:
					tt[coin] = 1
		for k in cls.trust_override:
			tt[k] = cls.trust_override[k]
		return tt

	trust_override = {'BTC':3,'BCH':3,'LTC':3,'DASH':1,'EMC':2}

	@classmethod
	def get_test_support(cls,coin,addr_type,network,toolname=None,verbose=False):
		"""
		If requested tool supports coin/addr_type/network triplet, return tool name.
		If 'tool' is None, return tool that supports coin/addr_type/network triplet.
		Return None on failure.
		"""
		all_tools = [toolname] if toolname else list(cls.external_tests[network].keys())
		coin = coin.upper()

		for tool in all_tools:
			if coin in cls.external_tests[network][tool]:
				break
		else:
			if verbose:
				m1 = 'Requested tool {t!r} does not support coin {c} on network {n}'
				m2 = 'No test tool found for coin {c} on network {n}'
				msg((m1 if toolname else m2).format(t=tool,c=coin,n=network))
			return None

		if addr_type == 'zcash_z':
			if toolname in (None,'zcash-mini'):
				return 'zcash-mini'
			else:
				if verbose:
					msg(f"Address type {addr_type!r} supported only by tool 'zcash-mini'")
				return None

		try:
			bl = cls.external_tests_blacklist[addr_type][tool]
		except:
			pass
		else:
			if bl is True or coin in bl:
				if verbose:
					msg(f'Tool {tool!r} blacklisted for coin {coin}, addr_type {addr_type!r}')
				return None

		if toolname: # skip whitelists
			return tool

		if addr_type in ('segwit','bech32'):
			st = cls.external_tests_segwit_whitelist
			if addr_type in st and coin in st[addr_type]:
				return tool
			else:
				if verbose:
					m1 = 'Requested tool {t!r} does not support coin {c}, addr_type {a!r}, on network {n}'
					m2 = 'No test tool found supporting coin {c}, addr_type {a!r}, on network {n}'
					msg((m1 if toolname else m2).format(t=tool,c=coin,n=network,a=addr_type))
				return None

		return tool

	external_tests = {
		'mainnet': {
			# List in order of preference.
			# If 'tool' is not specified, the first tool supporting the coin will be selected.
			'pycoin': (
				'DASH', # only compressed
				'BCH',
				'BTC','LTC','VIA','FTC','DOGE','MEC',
				'JBS','MZC','RIC','DFC','FAI','ARG','ZEC','DCR'),
			'keyconv': (
				'BCH',
				# broken: PIVX
				'42','AC','AIB','ANC','ARS','ATMOS','AUR','BLK','BQC','BTC','TEST','BTCD','CCC','CCN','CDN',
				'CLAM','CNC','CNOTE','CON','CRW','DEEPONION','DGB','DGC','DMD','DOGED','DOGE','DOPE',
				'DVC','EFL','EMC','EXCL','FAIR','FLOZ','FTC','GAME','GAP','GCR','GRC','GRS','GUN','HAM','HODL',
				'IXC','JBS','LBRY','LEAF','LTC','MMC','MONA','MUE','MYRIAD','MZC','NEOS','NLG','NMC','NVC',
				'NYAN','OK','OMC','PIGGY','PINK','PKB','PND','POT','PPC','PTC','PTS','QTUM','RBY','RDD',
				'RIC','SCA','SDC','SKC','SPR','START','SXC','TPC','UIS','UNO','VIA','VPN','VTC','WDC','WKC',
				'WUBS', 'XC', 'XPM', 'YAC', 'ZOOM', 'ZRC'),
			'ethkey': ('ETH','ETC'),
			'zcash-mini': ('ZEC',),
			'monero-python': ('XMR',),
		},
		'testnet': {
			'pycoin': {
				'DASH':'tDASH', # only compressed
				'BCH':'XTN',
				'BTC':'XTN','LTC':'XLT','VIA':'TVI','FTC':'FTX','DOGE':'XDT','DCR':'DCRT'
				},
			'ethkey': {},
			'keyconv': {}
		}
	}
	external_tests_segwit_whitelist = {
		# Whitelists apply to the *first* tool in cls.external_tests supporting the given coin/addr_type.
		# They're ignored if specific tool is requested.
		'segwit': ('BTC',), # LTC Segwit broken on pycoin: uses old fmt
		'bech32': ('BTC','LTC'),
		'compressed': (
			'BTC','LTC','VIA','FTC','DOGE','DASH','MEC','MYR','UNO',
			'JBS','MZC','RIC','DFC','FAI','ARG','ZEC','DCR','ZEC'
		),
	}
	external_tests_blacklist = {
		# Unconditionally block testing of the given coin/addr_type with given tool, or all coins if True
		'legacy': {},
		'segwit': { 'keyconv': True },
		'bech32': { 'keyconv': True },
	}

if __name__ == '__main__':

	opts_data = {
		'text': {
			'desc': 'Check altcoin data',
			'usage':'[opts]',
			'options': '-q, --quiet    Be quieter\n-v, --verbose  Be more verbose'
		}
	}

	cfg = Config( opts_data=opts_data, need_amt=False )

	msg('Checking CoinInfo WIF/P2PKH/P2SH version numbers and trust levels against protocol.py')
	TestCoinInfo.verify_core_coin_data( cfg, cfg.quiet, cfg.verbose )

	msg('Checking CoinInfo address leading symbols')
	TestCoinInfo.verify_leading_symbols( cfg.quiet, cfg.verbose )



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
scripts/create-token.py: Automated ERC20 token creation for the MMGen suite
"""

import sys,json,re
from subprocess import run,PIPE
from collections import namedtuple

import script_init
from mmgen.main import launch
from mmgen.cfg import Config
from mmgen.util import Msg,msg,rmsg,ymsg,die

ti = namedtuple('token_param_info',['default','conversion','test'])
class TokenData:
	fields = ('decimals','supply','name','symbol','owner_addr')
	decimals   = ti('18', int, lambda s: s.isascii() and s.isdigit() and 0 < int(s) <= 36)
	name       = ti(None, str, lambda s: s.isascii() and s.isprintable() and len(s) < 256)
	supply     = ti(None, int, lambda s: s.isascii() and s.isdigit() and 0 < int(s) < 2**256)
	symbol     = ti(None, str, lambda s: s.isascii() and s.isalnum() and len(s) <= 20)
	owner_addr = ti(None, str, lambda s: s.isascii() and s.isalnum() and len(s) == 40) # checked separately

token_data = TokenData()

req_solc_ver_pat = '^0.8.25'

opts_data = {
	'text': {
		'desc': 'Create an ERC20 token contract',
		'usage':'[opts] <owner address>',
		'options': f"""
-h, --help        Print this help message
-o, --outdir=D    Specify output directory for *.bin files
-d, --decimals=D  Number of decimals for the token (default: {token_data.decimals.default})
-n, --name=N      Token name (REQUIRED)
-p, --preprocess  Print the preprocessed code to stdout
-t, --supply=T    Total supply of the token (REQUIRED)
-s, --symbol=S    Token symbol (REQUIRED)
-S, --stdout      Output JSON data to stdout instead of files
-v, --verbose     Produce more verbose output
-c, --check-solc-version Check the installed solc version
""",
	'notes': """
The owner address must be in checksummed format.

Use ‘mmgen-tool eth_checksummed_addr’ to create it if necessary.
"""
	}
}

# ERC Token Standard #20 Interface
# https://github.com/ethereum/EIPs/blob/master/EIPS/eip-20-token-standard.md

solidity_code_template = """
// SPDX-License-Identifier: GPL-3.0-or-later

pragma solidity %s;

contract SafeMath {
    function safeAdd(uint a, uint b) public pure returns (uint c) {
        c = a + b;
        require(c >= a);
    }
    function safeSub(uint a, uint b) public pure returns (uint c) {
        require(b <= a);
        c = a - b;
    }
    function safeMul(uint a, uint b) public pure returns (uint c) {
        c = a * b;
        require(a == 0 || c / a == b);
    }
    function safeDiv(uint a, uint b) public pure returns (uint c) {
        require(b > 0);
        c = a / b;
    }
}

abstract contract ERC20Interface {
    function totalSupply() public virtual returns (uint);
    function balanceOf(address tokenOwner) public virtual returns (uint balance);
    function allowance(address tokenOwner, address spender) public virtual returns (uint remaining);
    function transfer(address to, uint tokens) public virtual returns (bool success);
    function approve(address spender, uint tokens) public virtual returns (bool success);
    function transferFrom(address from, address to, uint tokens) public virtual returns (bool success);

    event Transfer(address indexed from, address indexed to, uint tokens);
    event Approval(address indexed tokenOwner, address indexed spender, uint tokens);
}

contract Owned {
    address public owner;
    address public newOwner;

    event OwnershipTransferred(address indexed _from, address indexed _to);

    constructor() {
        owner = msg.sender;
    }

    modifier onlyOwner {
        require(msg.sender == owner);
        _;
    }

    function transferOwnership(address _newOwner) public onlyOwner {
        newOwner = _newOwner;
    }
    function acceptOwnership() public {
        require(msg.sender == newOwner);
        emit OwnershipTransferred(owner, newOwner);
        owner = newOwner;
        newOwner = address(0);
    }
}

// ----------------------------------------------------------------------------
// ERC20 Token, with the addition of symbol, name and decimals and assisted
// token transfers
// ----------------------------------------------------------------------------
contract Token is ERC20Interface, Owned, SafeMath {
    string public symbol;
    string public  name;
    uint8 public decimals;
    uint public _totalSupply;

    mapping(address => uint) balances;
    mapping(address => mapping(address => uint)) allowed;

    constructor() {
        symbol = "$symbol";
        name = "$name";
        decimals = $decimals;
        _totalSupply = $supply;
        balances[$owner_addr] = _totalSupply;
        emit Transfer(address(0), $owner_addr, _totalSupply);
    }
    function totalSupply() public view override returns (uint) {
        return _totalSupply  - balances[address(0)];
    }
    function balanceOf(address tokenOwner) public view override returns (uint balance) {
        return balances[tokenOwner];
    }
    function transfer(address to, uint tokens) public override returns (bool success) {
        balances[msg.sender] = safeSub(balances[msg.sender], tokens);
        balances[to] = safeAdd(balances[to], tokens);
        emit Transfer(msg.sender, to, tokens);
        return true;
    }
    function approve(address spender, uint tokens) public override returns (bool success) {
        allowed[msg.sender][spender] = tokens;
        emit Approval(msg.sender, spender, tokens);
        return true;
    }
    function transferFrom(address from, address to, uint tokens) public override returns (bool success) {
        balances[from] = safeSub(balances[from], tokens);
        allowed[from][msg.sender] = safeSub(allowed[from][msg.sender], tokens);
        balances[to] = safeAdd(balances[to], tokens);
        emit Transfer(from, to, tokens);
        return true;
    }
    function allowance(address tokenOwner, address spender) public view override returns (uint remaining) {
        return allowed[tokenOwner][spender];
    }
    // Owner can transfer out any accidentally sent ERC20 tokens
    function transferAnyERC20Token(address tokenAddress, uint tokens) public onlyOwner returns (bool success) {
        return ERC20Interface(tokenAddress).transfer(owner, tokens);
    }
}
""" % req_solc_ver_pat

def create_src(cfg,template,token_data):

	def gen():
		for k in token_data.fields:
			field = getattr(token_data,k)
			if k == 'owner_addr':
				owner_addr = cfg._args[0]
				from mmgen.addr import is_coin_addr
				if not is_coin_addr( cfg._proto, owner_addr.lower() ):
					die(1,f'{owner_addr}: not a valid {cfg._proto.coin} coin address')
				val = '0x' + owner_addr
			else:
				val = (
					getattr(cfg,k)
					or getattr(field,'default',None)
					or die(1,f'The --{k} option must be specified')
				)
				if not field.test(val):
					die(1,f'{val!r}: invalid parameter for option --{k}')

			yield (k, field.conversion(val))

	from string import Template
	return Template(template).substitute(**dict(gen()))

def check_solc_version():
	"""
	The output is used by other programs, so write to stdout only
	"""
	try:
		cp = run(['solc','--version'],check=True,stdout=PIPE)
	except:
		msg('solc missing or could not be executed') # this must go to stderr
		return False

	if cp.returncode != 0:
		Msg('solc exited with error')
		return False

	line = cp.stdout.decode().splitlines()[1]
	version_str = re.sub(r'Version:\s*','',line)
	m = re.match(r'(\d+)\.(\d+)\.(\d+)',version_str)

	if not m:
		Msg(f'Unrecognized solc version string: {version_str}')
		return False

	from semantic_version import Version,NpmSpec
	version = Version('{}.{}.{}'.format(*m.groups()))

	if version in NpmSpec(req_solc_ver_pat):
		Msg(str(version))
		return True
	else:
		Msg(f'solc version ({version_str}) does not match requirement ({req_solc_ver_pat})')
		return False

def compile_code(cfg,code):
	cmd = ['solc', '--optimize', '--bin', '--overwrite', '--evm-version=constantinople']
	if not cfg.stdout:
		cmd += ['--output-dir', cfg.outdir or '.']
	cmd += ['-']
	msg(f"Executing: {' '.join(cmd)}")
	cp = run(cmd,input=code.encode(),stdout=PIPE,stderr=PIPE)
	out = cp.stdout.decode().replace('\r','')
	err = cp.stderr.decode().replace('\r','').strip()
	if cp.returncode != 0:
		rmsg('Solidity compiler produced the following error:')
		msg(err)
		die(4,f'Solidity compiler exited with error (return val: {cp.returncode})')
	if err:
		ymsg('Solidity compiler produced the following warning:')
		msg(err)
	if cfg.stdout:
		o = out.split('\n')
		return {k:o[i+2] for k in ('SafeMath','Owned','Token') for i in range(len(o)) if k in o[i]}
	else:
		cfg._util.vmsg(out)

def main():
	cfg = Config(opts_data=opts_data)

	if cfg.check_solc_version:
		sys.exit(0 if check_solc_version() else 1)

	if not cfg._proto.coin in ('ETH','ETC'):
		die(1,'--coin option must be ETH or ETC')

	if not len(cfg._args) == 1:
		cfg._opts.usage()

	code = create_src( cfg, solidity_code_template, token_data )

	if cfg.preprocess:
		Msg(code)
		sys.exit(0)

	out = compile_code( cfg, code )

	if cfg.stdout:
		print(json.dumps(out))

	msg('Contract successfully compiled')

if __name__ == '__main__':

	launch(func=main)



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, a command-line cryptocurrency wallet
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tx.bump: transaction bump class
"""

from .new import New
from .completed import Completed
from ..util import msg,is_int,die

class Bump(Completed,New):
	desc = 'fee-bumped transaction'
	ext  = 'rawtx'
	bump_output_idx = None

	def __init__(self,check_sent,*args,**kwargs):

		super().__init__(*args,**kwargs)

		if not self.is_replaceable():
			die(1,f'Transaction {self.txid} is not replaceable')

		# If sending, require original tx to be sent
		if check_sent and not self.coin_txid:
			die(1,'Transaction {self.txid!r} was not broadcast to the network')

		self.coin_txid = ''
		self.sent_timestamp = None

	def check_sufficient_funds_for_bump(self):
		if not [o.amt for o in self.outputs if o.amt >= self.min_fee]:
			die(1,
				'Transaction cannot be bumped.\n' +
				f'All outputs contain less than the minimum fee ({self.min_fee} {self.coin})')

	def choose_output(self):

		def check_sufficient_funds(o_amt):
			if o_amt < self.min_fee:
				msg(f'Minimum fee ({self.min_fee} {self.coin}) is greater than output amount ({o_amt} {self.coin})')
				return False
			return True

		if len(self.outputs) == 1:
			if check_sufficient_funds(self.outputs[0].amt):
				self.bump_output_idx = 0
				return 0
			else:
				die(1,'Insufficient funds to bump transaction')

		init_reply = self.cfg.output_to_reduce
		chg_idx = self.chg_idx

		while True:
			if init_reply is None:
				from ..ui import line_input
				m = 'Choose an output to deduct the fee from (Hit ENTER for the change output): '
				reply = line_input( self.cfg, m ) or 'c'
			else:
				reply,init_reply = init_reply,None
			if chg_idx is None and not is_int(reply):
				msg('Output must be an integer')
			elif chg_idx is not None and not is_int(reply) and reply != 'c':
				msg("Output must be an integer, or 'c' for the change output")
			else:
				idx = chg_idx if reply == 'c' else (int(reply) - 1)
				if idx < 0 or idx >= len(self.outputs):
					msg(f'Output must be in the range 1-{len(self.outputs)}')
				else:
					o_amt = self.outputs[idx].amt
					cm = ' (change output)' if chg_idx == idx else ''
					prompt = f'Fee will be deducted from output {idx+1}{cm} ({o_amt} {self.coin})'
					if check_sufficient_funds(o_amt):
						if self.cfg.yes:
							msg(prompt)
						else:
							from ..ui import keypress_confirm
							if not keypress_confirm( self.cfg, prompt+'.  OK?', default_yes=True ):
								continue
						self.bump_output_idx = idx
						return idx



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
tw.unspent: Tracking wallet unspent outputs class for the MMGen suite
"""

from ..util import msg,suf,fmt
from ..obj import (
	ImmutableAttr,
	ListItemAttr,
	MMGenListItem,
	TwComment,
	HexStr,
	CoinTxID,
	NonNegativeInt )
from ..addr import CoinAddr
from .shared import TwMMGenID,get_tw_label
from .view import TwView

class TwUnspentOutputs(TwView):

	class display_type(TwView.display_type):

		class squeezed(TwView.display_type.squeezed):
			cols = ('num','txid','vout','addr','mmid','comment','amt','amt2','date')

		class detail(TwView.display_type.detail):
			cols = ('num','txid','vout','addr','mmid','amt','amt2','block','date_time','comment')

	show_mmid = True
	no_rpcdata_errmsg = """
		No spendable outputs found!  Import addresses with balances into your
		watch-only wallet using 'mmgen-addrimport' and then re-run this program.
	"""
	update_widths_on_age_toggle = False
	print_output_types = ('detail',)
	mod_subpath = 'tw.unspent'

	class MMGenTwUnspentOutput(MMGenListItem):
		txid         = ListItemAttr(CoinTxID)
		vout         = ListItemAttr(NonNegativeInt)
		amt          = ImmutableAttr(None)
		amt2         = ListItemAttr(None) # the ETH balance for token account
		comment      = ListItemAttr(TwComment,reassign_ok=True)
		twmmid       = ImmutableAttr(TwMMGenID,include_proto=True)
		addr         = ImmutableAttr(CoinAddr,include_proto=True)
		confs        = ImmutableAttr(int,typeconv=False)
		date         = ListItemAttr(int,typeconv=False,reassign_ok=True)
		scriptPubKey = ImmutableAttr(HexStr)
		skip         = ListItemAttr(str,typeconv=False,reassign_ok=True)

		def __init__(self,proto,**kwargs):
			self.__dict__['proto'] = proto
			MMGenListItem.__init__(self,**kwargs)

		class conv_funcs:
			@staticmethod
			def amt(instance,value):
				return instance.proto.coin_amt(value)
			@staticmethod
			def amt2(instance,value):
				return instance.proto.coin_amt(value)

	async def __init__(self,cfg,proto,minconf=1,addrs=[]):
		await super().__init__(cfg,proto)
		self.minconf  = minconf
		self.addrs    = addrs
		from ..cfg import gc
		self.min_cols = gc.min_screen_width

	@property
	def total(self):
		return sum(i.amt for i in self.data)

	def gen_data(self,rpc_data,lbl_id):
		for o in rpc_data:
			if not lbl_id in o:
				continue # coinbase outputs have no account field
			l = get_tw_label(self.proto,o[lbl_id])
			if l:
				o.update({
					'twmmid':  l.mmid,
					'comment': l.comment or '',
					'amt':     self.proto.coin_amt(o['amount']),
					'addr':    CoinAddr(self.proto,o['address']),
					'confs':   o['confirmations']
				})
				yield self.MMGenTwUnspentOutput(
					self.proto,
					**{ k:v for k,v in o.items() if k in self.MMGenTwUnspentOutput.valid_attrs } )

	def filter_data(self):

		data = self.data.copy()

		for d in data:
			d.skip = ''

		gkeys = {'addr':'addr','twmmid':'addr','txid':'txid'}
		if self.group and self.sort_key in gkeys:
			for a,b in [(data[i],data[i+1]) for i in range(len(data)-1)]:
				for k in gkeys:
					if self.sort_key == k and getattr(a,k) == getattr(b,k):
						b.skip = gkeys[k]

		return data

	def get_column_widths(self,data,wide,interactive):

		show_mmid = self.show_mmid or wide

		# num txid vout addr [mmid] [comment] amt [amt2] date
		return self.compute_column_widths(
			widths = { # fixed cols
				'num': max(2,len(str(len(data)))+1),
				'vout': 4,
				'mmid': max(len(d.twmmid.disp) for d in data) if show_mmid else 0,
				'amt': self.amt_widths['amt'],
				'amt2': self.amt_widths.get('amt2',0),
				'block': self.age_col_params['block'][0] if wide else 0,
				'date_time': self.age_col_params['date_time'][0] if wide else 0,
				'date': self.age_w,
				'spc': 7 if show_mmid else 5, # 7(5) spaces in fs
			},
			maxws = { # expandable cols
				'txid': self.txid_w,
				'addr': max(len(d.addr) for d in data),
				'comment': max(d.comment.screen_width for d in data) if show_mmid else 0,
			},
			minws = {
				'txid': 7,
				'addr': 10,
				'comment': len('Comment') if show_mmid else 0,
			},
			maxws_nice = {'txid':12, 'addr':16} if show_mmid else {'txid':12},
			wide = wide,
			interactive = interactive,
		)

	def squeezed_col_hdr(self,cw,fs,color):
		return fs.format(
			n = '',
			t = 'TxID',
			v = 'Vout',
			a = 'Address',
			m = 'MMGenID',
			c = 'Comment',
			A = 'Amt({})'.format(self.proto.dcoin),
			B = 'Amt({})'.format(self.proto.coin),
			d = self.age_hdr )

	def detail_col_hdr(self,cw,fs,color):
		return fs.format(
			n = '',
			t = 'TxID',
			v = 'Vout',
			a = 'Address',
			m = 'MMGenID',
			A = 'Amt({})'.format(self.proto.dcoin),
			B = 'Amt({})'.format(self.proto.coin),
			b = 'Block',
			D = 'Date/Time',
			c = 'Comment' )

	def gen_squeezed_display(self,data,cw,fs,color,fmt_method):

		for n,d in enumerate(data):
			yield fs.format(
				n = str(n+1) + ')',
				t = (d.txid.fmtc( '|' + '.'*(cw.txid-1), width=cw.txid, color=color ) if d.skip  == 'txid'
					else d.txid.truncate( width=cw.txid, color=color )) if cw.txid else None,
				v = ' ' + d.vout.fmt( width=cw.vout-1, color=color ) if cw.vout else None,
				a = d.addr.fmtc( '|' + '.'*(cw.addr-1), width=cw.addr, color=color ) if d.skip == 'addr'
					else d.addr.fmt( width=cw.addr, color=color ),
				m = (d.twmmid.fmtc( '.'*cw.mmid, width=cw.mmid, color=color ) if d.skip == 'addr'
					else d.twmmid.fmt( width=cw.mmid, color=color )) if cw.mmid else None,
				c = d.comment.fmt2( width=cw.comment, color=color, nullrepl='-' ) if cw.comment else None,
				A = d.amt.fmt( color=color, iwidth=cw.iwidth, prec=self.disp_prec ),
				B = d.amt2.fmt( color=color, iwidth=cw.iwidth2, prec=self.disp_prec ) if cw.amt2 else None,
				d = self.age_disp(d,self.age_fmt),
			)

	def gen_detail_display(self,data,cw,fs,color,fmt_method):

		for n,d in enumerate(data):
			yield fs.format(
				n = str(n+1) + ')',
				t = d.txid.fmt( width=cw.txid, color=color ) if cw.txid else None,
				v = ' ' + d.vout.fmt( width=cw.vout-1, color=color ) if cw.vout else None,
				a = d.addr.fmt( width=cw.addr, color=color ),
				m = d.twmmid.fmt( width=cw.mmid, color=color ),
				A = d.amt.fmt( color=color, iwidth=cw.iwidth, prec=self.disp_prec ),
				B = d.amt2.fmt( color=color, iwidth=cw.iwidth2, prec=self.disp_prec ) if cw.amt2 else None,
				b = self.age_disp(d,'block'),
				D = self.age_disp(d,'date_time'),
				c = d.comment.fmt2( width=cw.comment, color=color, nullrepl='-' ))

	def display_total(self):
		msg('\nTotal unspent: {} {} ({} output{})'.format(
			self.total.hl(),
			self.proto.dcoin,
			len(self.data),
			suf(self.data) ))

	async def set_dates(self,us):
		if not self.dates_set:
			# 'blocktime' differs from 'time', is same as getblockheader['time']
			dates = [ o.get('blocktime',0)
				for o in await self.rpc.gathered_icall('gettransaction',[(o.txid,True,False) for o in us]) ]
			for idx,o in enumerate(us):
				o.date = dates[idx]
			self.dates_set = True

	class sort_action(TwView.sort_action):

		def s_twmmid(self,parent):
			parent.do_sort('twmmid')
			parent.show_mmid = True

	class display_action(TwView.display_action):

		def d_mmid(self,parent):
			parent.show_mmid = not parent.show_mmid

		def d_group(self,parent):
			if parent.can_group:
				parent.group = not parent.group



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
main_wallet: Entry point for MMGen wallet-related scripts
"""

import sys,os
from .cfg import gc,Config
from .color import green,yellow
from .util import msg,gmsg_r,ymsg,bmsg,die,capfirst
from .wallet import Wallet,get_wallet_cls

usage = '[opts] [infile]'
nargs = 1
iaction = 'convert'
oaction = 'convert'
do_bw_note = True
do_sw_note = False
do_ss_note = False

invoked_as = {
	'mmgen-walletgen':    'gen',
	'mmgen-walletconv':   'conv',
	'mmgen-walletchk':    'chk',
	'mmgen-passchg':      'passchg',
	'mmgen-subwalletgen': 'subgen',
	'mmgen-seedsplit':    'seedsplit',
}[gc.prog_name]

dsw = f'the default or specified {gc.proj_name} wallet'

# full: defhHiJkKlLmoOpPqrSvz-
if invoked_as == 'gen':
	desc = f'Generate an {gc.proj_name} wallet from a random seed'
	opt_filter = 'ehdoJlLpPqrSvz-'
	usage = '[opts]'
	oaction = 'output'
	nargs = 0
elif invoked_as == 'conv':
	desc = 'Convert ' + dsw + ' from one format to another'
	opt_filter = 'dehHiJkKlLmNoOpPqrSvz-'
elif invoked_as == 'chk':
	desc = 'Check validity of ' + dsw
	opt_filter = 'ehiHOlpPqrvz-'
	iaction = 'input'
elif invoked_as == 'passchg':
	desc = 'Change the passphrase, hash preset or label of ' + dsw
	opt_filter = 'efhdiHkKOlLmNpPqrSvz-'
	iaction = 'input'
	do_bw_note = False
elif invoked_as == 'subgen':
	desc = 'Generate a subwallet from ' + dsw
	opt_filter = 'dehHiJkKlLmNoOpPqrSvz-' # omitted: f
	usage = '[opts] [infile] <Subseed Index>'
	iaction = 'input'
	oaction = 'output'
	do_sw_note = True
elif invoked_as == 'seedsplit':
	desc = 'Generate a seed share from ' + dsw
	opt_filter = 'dehHiJlLMNIoOpPqrSvz-'
	usage = '[opts] [infile] [<Split ID String>:]<index>:<share count>'
	iaction = 'input'
	oaction = 'output'
	do_ss_note = True

opts_data = {
	'text': {
		'desc': desc,
		'usage': usage,
		'options': """
-h, --help            Print this help message
--, --longhelp        Print help message for long options (common options)
-d, --outdir=      d  Output files to directory 'd' instead of working dir
-e, --echo-passphrase Echo passphrases and other user input to screen
-f, --force-update    Force update of wallet even if nothing has changed
-i, --in-fmt=      f  {iaction} from wallet format 'f' (see FMT CODES below)
-o, --out-fmt=     f  {oaction} to wallet format 'f' (see FMT CODES below)
-H, --hidden-incog-input-params=f,o  Read hidden incognito data from file
                      'f' at offset 'o' (comma-separated)
-J, --hidden-incog-output-params=f,o  Write hidden incognito data to file
                      'f' at offset 'o' (comma-separated). File 'f' will be
                      created if necessary and filled with random data.
-O, --old-incog-fmt   Specify old-format incognito input
-k, --keep-passphrase Reuse passphrase of input wallet for output wallet
-K, --keep-hash-preset Reuse hash preset of input wallet for output wallet
-l, --seed-len=    l  Specify wallet seed length of 'l' bits.  This option
                      is required only for brainwallet and incognito inputs
                      with non-standard (< {dsl}-bit) seed lengths.
-L, --label=       l  Specify a label 'l' for output wallet
-m, --keep-label      Reuse label of input wallet for output wallet
-M, --master-share=i  Use a master share with index 'i' (min:{ms_min}, max:{ms_max})
-p, --hash-preset= p  Use the scrypt hash parameters defined by preset 'p'
                      for password hashing (default: '{gc.dfl_hash_preset}')
-z, --show-hash-presets Show information on available hash presets
-P, --passwd-file= f  Get wallet passphrase from file 'f'
-N, --passwd-file-new-only Use passwd file only for new, not existing, wallet
-q, --quiet           Produce quieter output; suppress some warnings
-r, --usr-randchars=n Get 'n' characters of additional randomness from user
                      (min={cfg.min_urandchars}, max={cfg.max_urandchars}, default={cfg.usr_randchars})
-S, --stdout          Write wallet data to stdout instead of file
-v, --verbose         Produce more verbose output
""",
	'notes': """

{n_ss}{n_sw}{n_pw}{n_bw}

FMT CODES:

  {f}
"""
	},
	'code': {
		'options': lambda cfg,help_notes,s: s.format(
			iaction=capfirst(iaction),
			oaction=capfirst(oaction),
			ms_min=help_notes('MasterShareIdx').min_val,
			ms_max=help_notes('MasterShareIdx').max_val,
			dsl=help_notes('dfl_seed_len'),
			cfg=cfg,
			gc=gc,
		),
		'notes': lambda cfg,help_mod,help_notes,s: s.format(
			f=help_notes('fmt_codes'),
			n_ss=('',help_mod('seedsplit')+'\n\n')[do_ss_note],
			n_sw=('',help_notes('subwallet')+'\n\n')[do_sw_note],
			n_pw=help_notes('passwd'),
			n_bw=('','\n\n'+help_notes('brainwallet'))[do_bw_note]
		)
	}
}

cfg = Config( opts_data=opts_data, opt_filter=opt_filter, need_proto=False )

cmd_args = cfg._args

if invoked_as == 'subgen':
	from .subseed import SubSeedIdx
	ss_idx = SubSeedIdx(cmd_args.pop())
elif invoked_as == 'seedsplit':
	from .obj import get_obj
	from .seedsplit import SeedSplitSpecifier,MasterShareIdx
	master_share = MasterShareIdx(cfg.master_share) if cfg.master_share else None
	if cmd_args:
		sss = get_obj(SeedSplitSpecifier,s=cmd_args.pop(),silent=True)
		if master_share:
			if not sss:
				sss = SeedSplitSpecifier('1:2')
			elif sss.idx == 1:
				m1 = 'Share index of 1 meaningless in master share context.'
				m2 = 'To generate a master share, omit the seed split specifier.'
				die(1,m1+'  '+m2)
		elif not sss:
			cfg._opts.usage()
	elif master_share:
		sss = SeedSplitSpecifier('1:2')
	else:
		cfg._opts.usage()

from .fileutil import check_infile,get_seed_file

if cmd_args:
	if invoked_as == 'gen' or len(cmd_args) > 1:
		cfg._opts.usage()
	check_infile(cmd_args[0])

sf = get_seed_file(cfg,nargs,invoked_as=invoked_as)

if invoked_as != 'chk':
	from .ui import do_license_msg
	do_license_msg(cfg)

if invoked_as == 'gen':
	ss_in = None
else:
	ss_in = Wallet(
		cfg     = cfg,
		fn      = sf,
		passchg = invoked_as == 'passchg',
		passwd_file = False if cfg.passwd_file_new_only else None )
	m1 = green('Processing input wallet ')
	m2 = ss_in.seed.sid.hl()
	m3 = yellow(' (default wallet)') if sf and os.path.dirname(sf) == cfg.data_dir else ''
	msg(m1+m2+m3)

if invoked_as == 'chk':
	lbl = ss_in.ssdata.label.hl() if hasattr(ss_in.ssdata,'label') else 'NONE'
	cfg._util.vmsg(f'Wallet label: {lbl}')
	# TODO: display creation date
	sys.exit(0)

if invoked_as != 'gen':
	gmsg_r('Processing output wallet' + ('\n',' ')[invoked_as == 'seedsplit'])

if invoked_as == 'subgen':
	ss_out = Wallet(
		cfg      = cfg,
		seed_bin = ss_in.seed.subseed(ss_idx,print_msg=True).data )
elif invoked_as == 'seedsplit':
	shares = ss_in.seed.split(sss.count,sss.id,master_share)
	seed_out = shares.get_share_by_idx(sss.idx,base_seed=True)
	msg(seed_out.get_desc(ui=True))
	ss_out = Wallet(
		cfg  = cfg,
		seed = seed_out )
else:
	ss_out = Wallet(
		cfg     = cfg,
		ss      = ss_in,
		passchg = invoked_as == 'passchg' )

if invoked_as == 'gen':
	cfg._util.qmsg(f"This wallet's Seed ID: {ss_out.seed.sid.hl()}")

if invoked_as == 'passchg':
	def data_changed(attrs):
		for attr in attrs:
			if getattr(ss_out.ssdata,attr) != getattr(ss_in.ssdata,attr):
				return True
		return False
	if not ( cfg.force_update or data_changed(('passwd','hash_preset','label')) ):
		die(1,'Password, hash preset and label are unchanged.  Taking no action')

if invoked_as == 'passchg':

	def secure_delete(fn):
		bmsg('Securely deleting old wallet')
		from .fileutil import shred_file
		shred_file( fn, verbose = cfg.verbose )

	def rename_old_wallet_maybe(silent):
		# though very unlikely, old and new wallets could have same Key ID and thus same filename.
		# If so, rename old wallet file before deleting.
		old_fn = ss_in.infile.name
		if os.path.basename(old_fn) == ss_out._filename():
			if not silent:
				ymsg(
					'Warning: diverting old wallet {old_fn!r} due to Key ID collision.  ' +
					'Please securely delete or move the diverted file!' )
			os.rename( old_fn, old_fn+'.divert' )
			return old_fn+'.divert'
		else:
			return old_fn

	if ss_in.infile.dirname == cfg.data_dir:
		from .ui import confirm_or_raise
		confirm_or_raise(
			cfg      = cfg,
			message  = yellow('Confirmation of default wallet update'),
			action   = 'update the default wallet',
			exit_msg = 'Password not changed' )
		old_wallet = rename_old_wallet_maybe(silent=True)
		ss_out.write_to_file( desc='New wallet', outdir=cfg.data_dir )
		secure_delete( old_wallet )
	else:
		old_wallet = rename_old_wallet_maybe(silent=False)
		ss_out.write_to_file()
		from .ui import keypress_confirm
		if keypress_confirm( cfg, f'Securely delete old wallet {old_wallet!r}?' ):
			secure_delete( old_wallet )
elif invoked_as == 'gen' and not cfg.outdir and not cfg.stdout:
	from .filename import find_file_in_dir
	if find_file_in_dir( get_wallet_cls('mmgen'), cfg.data_dir ):
		ss_out.write_to_file()
	else:
		from .ui import keypress_confirm
		if keypress_confirm(
				cfg,
				'Make this wallet your default and move it to the data directory?',
				default_yes = True ):
			ss_out.write_to_file(outdir=cfg.data_dir)
		else:
			ss_out.write_to_file()
else:
	ss_out.write_to_file()

if invoked_as == 'passchg':
	if ss_out.ssdata.passwd == ss_in.ssdata.passwd:
		msg('New and old passphrases are the same')
	else:
		msg('Wallet passphrase has changed')
	if ss_out.ssdata.hash_preset != ss_in.ssdata.hash_preset:
		msg(f'Hash preset has been changed to {ss_out.ssdata.hash_preset!r}')



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, a command-line cryptocurrency wallet
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
devinit: Developer tools initialization for the MMGen suite
"""

devtools_funcs = {
	'pfmt':              lambda *args,**kwargs: devtools_call('pfmt',*args,**kwargs),
	'pmsg':              lambda *args,**kwargs: devtools_call('pmsg',*args,**kwargs),
	'pmsg_r':            lambda *args,**kwargs: devtools_call('pmsg_r',*args,**kwargs),
	'pdie':              lambda *args,**kwargs: devtools_call('pdie',*args,**kwargs),
	'pexit':             lambda *args,**kwargs: devtools_call('pexit',*args,**kwargs),
	'Pmsg':              lambda *args,**kwargs: devtools_call('Pmsg',*args,**kwargs),
	'Pdie':              lambda *args,**kwargs: devtools_call('Pdie',*args,**kwargs),
	'Pexit':             lambda *args,**kwargs: devtools_call('Pexit',*args,**kwargs),
	'print_stack_trace': lambda *args,**kwargs: devtools_call('print_stack_trace',*args,**kwargs),
	'get_diff':          lambda *args,**kwargs: devtools_call('get_diff',*args,**kwargs),
	'print_diff':        lambda *args,**kwargs: devtools_call('print_diff',*args,**kwargs),
	'get_ndiff':         lambda *args,**kwargs: devtools_call('get_ndiff',*args,**kwargs),
	'print_ndiff':       lambda *args,**kwargs: devtools_call('print_ndiff',*args,**kwargs),
}

def devtools_call(funcname,*args,**kwargs):
	from . import devtools
	return getattr(devtools,funcname)(*args,**kwargs)

def MMGenObject_call(methodname,*args,**kwargs):
	from .devtools import MMGenObjectMethods
	return getattr(MMGenObjectMethods,methodname)(*args,**kwargs)

class MMGenObjectDevTools:

	pmsg  = lambda *args,**kwargs: MMGenObject_call('pmsg',*args,**kwargs)
	pdie  = lambda *args,**kwargs: MMGenObject_call('pdie',*args,**kwargs)
	pexit = lambda *args,**kwargs: MMGenObject_call('pexit',*args,**kwargs)
	pfmt  = lambda *args,**kwargs: MMGenObject_call('pfmt',*args,**kwargs)

	# Check that all immutables have been initialized.  Expensive, so do only when testing.
	def immutable_attr_init_check(self):

		cls = type(self)

		for attrname in self.valid_attrs:

			for o in (cls,cls.__bases__[0]): # assume there's only one base class
				if attrname in o.__dict__:
					attr = o.__dict__[attrname]
					break
			else:
				from .util import die
				die(4,f'unable to find descriptor {cls.__name__}.{attrname}')

			if type(attr).__name__ == 'ImmutableAttr' and attrname not in self.__dict__:
				from .util import die
				die(4,f'attribute {attrname!r} of {cls.__name__} has not been initialized in constructor!')

def init_dev():
	import builtins
	# MMGenObject is added to the namespace by objmethods.py, so we must name the builtin differently
	# to avoid inadvertently adding MMGenObject to the global namespace here:
	setattr(builtins,'MMGenObjectDevTools',MMGenObjectDevTools)
	for funcname,func in devtools_funcs.items():
		setattr(builtins,funcname,func)



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>

"""
test.objtest_py_d.ot_common: shared data for MMGen data objects tests
"""

from ..include.common import getrand

r16,r32 = getrand(16),getrand(32)



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, a command-line cryptocurrency wallet
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tw.shared: classes and functions shared by all tracking wallet classes
"""

from ..objmethods import HiliteStr,InitErrors,MMGenObject
from ..obj import TwComment
from ..addr import MMGenID

class TwMMGenID(HiliteStr,InitErrors,MMGenObject):
	color = 'orange'
	width = 0
	trunc_ok = False
	def __new__(cls,proto,id_str):
		if isinstance(id_str,cls):
			return id_str
		try:
			ret = addr = disp = MMGenID(proto,id_str)
			sort_key,idtype = (ret.sort_key,'mmgen')
		except Exception as e:
			try:
				coin,addr = id_str.split(':',1)
				assert coin == proto.base_coin.lower(),(
					f'not a string beginning with the prefix {proto.base_coin.lower()!r}:' )
				assert addr.isascii() and addr.isalnum(), 'not an ASCII alphanumeric string'
				ret,sort_key,idtype,disp = (id_str,'z_'+id_str,'non-mmgen','non-MMGen')
				addr = proto.coin_addr(addr)
			except Exception as e2:
				return cls.init_fail(e,id_str,e2=e2)

		me = str.__new__(cls,ret)
		me.obj = ret
		me.disp = disp
		me.addr = addr
		me.sort_key = sort_key
		me.type = idtype
		me.proto = proto
		return me

	def fmt(self,**kwargs):
		return super().fmtc(self.disp,**kwargs)

# non-displaying container for TwMMGenID,TwComment
class TwLabel(str,InitErrors,MMGenObject):
	exc = 'BadTwLabel'
	passthru_excs = ('BadTwComment',)
	def __new__(cls,proto,text):
		if isinstance(text,cls):
			return text
		try:
			ts = text.split(None,1)
			mmid = TwMMGenID(proto,ts[0])
			comment = TwComment(ts[1] if len(ts) == 2 else '')
			me = str.__new__( cls, mmid + (' ' + comment if comment else '') )
			me.mmid = mmid
			me.comment = comment
			me.proto = proto
			return me
		except Exception as e:
			return cls.init_fail(e,text)

def get_tw_label(proto,s):
	"""
	raise an exception on a malformed comment, return None on an empty or invalid label
	"""
	try:
		return TwLabel(proto,s)
	except Exception as e:
		if type(e).__name__ == 'BadTwComment': # do it this way to avoid importing .exception
			raise
		else:
			return None



#!/usr/bin/env python3

import os

from mmgen.cfg import Config
from mmgen.util import msg

cfg = Config(process_opts=True)

cmd_args = cfg._args

from mmgen.cfgfile import mmgen_cfg_file

cf_usr = mmgen_cfg_file(cfg,'usr')
cf_sys = mmgen_cfg_file(cfg,'sys')
cf_sample = mmgen_cfg_file(cfg,'sample')

msg(f'Usr cfg file:    {os.path.relpath(cf_usr.fn)}')
msg(f'Sys cfg file:    {os.path.relpath(cf_sys.fn)}')
msg(f'Sample cfg file: {os.path.relpath(cf_sample.fn)}')

if cmd_args:
	if cmd_args[0] == 'parse_test':
		ps = cf_sample.get_lines()
		msg(f'parsed chunks: {len(ps)}')
		pu = cf_usr.get_lines()
		msg('usr cfg: {}'.format( ' '.join(f'{i.name}={i.value}' for i in pu) ))
	elif cmd_args[0] == 'coin_specific_vars':
		for varname in cmd_args[1:]:
			msg('{}.{}: {}'.format(
				type(cfg._proto).__name__,
				varname,
				getattr(cfg._proto,varname)
			))
	elif cmd_args[0] == 'autoset_opts':
		assert cfg.rpc_backend == 'aiohttp', "cfg.rpc_backend != 'aiohttp'"
	elif cmd_args[0] == 'autoset_opts_cmdline':
		assert cfg.rpc_backend == 'curl', "cfg.rpc_backend != 'curl'"
	elif cmd_args[0] == 'mnemonic_entry_modes':
		from mmgen.mn_entry import mn_entry
		msg('mnemonic_entry_modes: {}\nmmgen: {}\nbip39: {}'.format(
			cfg.mnemonic_entry_modes,
			mn_entry(cfg,'mmgen').usr_dfl_entry_mode,
			mn_entry(cfg,'bip39').usr_dfl_entry_mode ))



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test.cmdtest_py_d.ct_ethdev: Ethdev tests for the cmdtest.py test suite
"""

import sys,os,re,shutil,asyncio,json
from decimal import Decimal
from collections import namedtuple
from subprocess import run,PIPE,DEVNULL
from pathlib import Path

from mmgen.color import yellow,blue,cyan,set_vt100
from mmgen.util import msg,rmsg,die

from ..include.common import (
	cfg,
	check_solc_ver,
	omsg,
	imsg,
	imsg_r,
	joinpath,
	read_from_file,
	write_to_file,
	cmp_or_die,
	strip_ansi_escapes,
	silence,
	end_silence,
	gr_uc,
	stop_test_daemons
)
from .common import (
	ref_dir,
	dfl_words_file,
	tx_comment_jp,
	tx_comment_lat_cyr_gr,
	tw_comment_zh,
	tw_comment_lat_cyr_gr,
	get_file_with_ext,
	ok_msg,
	Ctrl_U
)
from .ct_base import CmdTestBase
from .ct_shared import CmdTestShared

del_addrs = ('4','1')
dfl_sid = '98831F3A'

# The OpenEthereum dev address with lots of coins.  Create with "ethkey -b info ''":
dfl_devaddr = '00a329c0648769a73afac7f9381e08fb43dbea72'
dfl_devkey = '4d5db4107d237df6a3d58ee5f70ae63d73d7658d4026f2eefd2f204c81682cb7'

burn_addr  = 'deadbeef'*5
burn_addr2 = 'beadcafe'*5

amt1 = '999999.12345689012345678'
amt2 = '888.111122223333444455'

parity_devkey_fn = 'parity.devkey'

def set_vbals(daemon_id):
	global vbal1, vbal2, vbal3, vbal4, vbal5, vbal6, vbal7, vbal9
	if daemon_id == 'geth':
		vbal1 = '1.2288334'
		vbal2 = '99.996560752'
		vbal3 = '1.2314176'
		vbal4 = '127.0287834'
		vbal5 = '1000126.14775104212345678'
		vbal6 = '1000126.14880104212345678'
		vbal7 = '1000124.91891764212345678'
		vbal9 = '1.2262504'
	else:
		vbal1 = '1.2288409'
		vbal2 = '99.997088755'
		vbal3 = '1.23142525'
		vbal4 = '127.0287909'
		vbal5 = '1000126.14828654512345678'
		vbal6 = '1000126.14933654512345678'
		vbal7 = '1000124.91944564512345678'
		vbal9 = '1.22626295'

bals = lambda k: {
	'1': [  ('98831F3A:E:1','123.456')],
	'2': [  ('98831F3A:E:1','123.456'),('98831F3A:E:11','1.234')],
	'3': [  ('98831F3A:E:1','123.456'),('98831F3A:E:11','1.234'),('98831F3A:E:21','2.345')],
	'4': [  ('98831F3A:E:1','100'),
			('98831F3A:E:2','23.45495'),
			('98831F3A:E:11','1.234'),
			('98831F3A:E:21','2.345')],
	'5': [  ('98831F3A:E:1','100'),
			('98831F3A:E:2','23.45495'),
			('98831F3A:E:11','1.234'),
			('98831F3A:E:21','2.345'),
			(burn_addr + r'\s+non-MMGen',amt1)],
	'8': [  ('98831F3A:E:1','0'),
			('98831F3A:E:2','23.45495'),
			('98831F3A:E:11',vbal1,'a1'),
			('98831F3A:E:12','99.99895'),
			('98831F3A:E:21','2.345'),
			(burn_addr + r'\s+non-MMGen',amt1)],
	'9': [  ('98831F3A:E:1','0'),
			('98831F3A:E:2','23.45495'),
			('98831F3A:E:11',vbal1,'a1'),
			('98831F3A:E:12',vbal2),
			('98831F3A:E:21','2.345'),
			(burn_addr + r'\s+non-MMGen',amt1)],
	'10': [ ('98831F3A:E:1','0'),
			('98831F3A:E:2','23.0218'),
			('98831F3A:E:3','0.4321'),
			('98831F3A:E:11',vbal1,'a1'),
			('98831F3A:E:12',vbal2),
			('98831F3A:E:21','2.345'),
			(burn_addr + r'\s+non-MMGen',amt1)]
}[k]

token_bals = lambda k: {
	'1': [  ('98831F3A:E:11','1000','1.234')],
	'2': [  ('98831F3A:E:11','998.76544',vbal3,'a1'),
			('98831F3A:E:12','1.23456','0')],
	'3': [  ('98831F3A:E:11','110.654317776666555545',vbal1,'a1'),
			('98831F3A:E:12','1.23456','0')],
	'4': [  ('98831F3A:E:11','110.654317776666555545',vbal1,'a1'),
			('98831F3A:E:12','1.23456','0'),
			(burn_addr + r'\s+non-MMGen',amt2,amt1)],
	'5': [  ('98831F3A:E:11','110.654317776666555545',vbal1,'a1'),
			('98831F3A:E:12','1.23456','99.99895'),
			(burn_addr + r'\s+non-MMGen',amt2,amt1)],
	'6': [  ('98831F3A:E:11','110.654317776666555545',vbal1,'a1'),
			('98831F3A:E:12','0',vbal2),
			('98831F3A:E:13','1.23456','0'),
			(burn_addr + r'\s+non-MMGen',amt2,amt1)],
	'7': [  ('98831F3A:E:11','67.444317776666555545',vbal9,'a2'),
			('98831F3A:E:12','43.21',vbal2),
			('98831F3A:E:13','1.23456','0'),
			(burn_addr + r'\s+non-MMGen',amt2,amt1)]
}[k]

token_bals_getbalance = lambda k: {
	'1': (vbal4,'999999.12345689012345678'),
	'2': ('111.888877776666555545','888.111122223333444455')
}[k]

coin = cfg.coin

class CmdTestEthdev(CmdTestBase,CmdTestShared):
	'Ethereum transacting, token deployment and tracking wallet operations'
	networks = ('eth','etc')
	passthru_opts = ('coin','daemon_id','http_timeout','rpc_backend')
	extra_spawn_args = ['--regtest=1']
	tmpdir_nums = [22]
	color = True
	cmd_group_in = (
		('setup',             f'dev mode tests for coin {coin} (start daemon)'),
		('subgroup.misc',     []),
		('subgroup.init',     []),
		('subgroup.msg',      ['init']),
		('subgroup.main',     ['init']),
		('subgroup.contract', ['main']),
		('subgroup.token',    ['contract']),
		('subgroup.twexport', ['token']),
		('subgroup.cached',   ['token']),
		('subgroup.view',     ['cached']),
		('subgroup.label',    ['cached']),
		('subgroup.remove',   ['cached']),
		('stop',              'stopping daemon'),
	)
	cmd_subgroups = {
	'misc': (
		'miscellaneous commands',
		('daemon_version', 'mmgen-tool daemon_version'),
	),
	'init': (
		'initializing wallets',
		('wallet_upgrade1',     'upgrading the tracking wallet (v1 -> v2)'),
		('wallet_upgrade2',     'upgrading the tracking wallet (v2 -> v3)'),
		('addrgen',             'generating addresses'),
		('addrimport',          'importing addresses'),
		('addrimport_dev_addr', "importing dev faucet address 'Ox00a329c..'"),
		('fund_dev_address',    'funding the default (Parity dev) address'),
	),
	'msg': (
		'message signing',
		('msgsign_chk',          "signing a message (low-level, check against 'eth_sign' RPC call)"),
		('msgcreate',            'creating a message file'),
		('msgsign',              'signing the message file'),
		('msgverify',            'verifying the message file'),
		('msgexport',            'exporting the message file data to JSON for third-party verifier'),
		('msgverify_export',     'verifying the exported JSON data'),

		('msgcreate_raw',        'creating a message file (--msghash-type=raw)'),
		('msgsign_raw',          'signing the message file (msghash_type=raw)'),
		('msgverify_raw',        'verifying the message file (msghash_type=raw)'),
		('msgexport_raw',        'exporting the message file data to JSON (msghash_type=raw)'),
		('msgverify_export_raw', 'verifying the exported JSON data (msghash_type=raw)'),
	),
	'main': (
		'creating, signing, sending and bumping Ethereum transactions',
		('txcreate1',            'creating a transaction (spend from dev address to address :1)'),
		('txview1_raw',          'viewing the raw transaction'),
		('txsign1',              'signing the transaction'),
		('txview1_sig',          'viewing the signed transaction'),
		('tx_status0_bad',       'getting the transaction status'),
		('txsign1_ni',           'signing the transaction (non-interactive)'),
		('txsend1',              'sending the transaction'),
		('bal1',                 f'the {coin} balance'),

		('txcreate2',            'creating a transaction (spend from dev address to address :11)'),
		('txsign2',              'signing the transaction'),
		('txsend2',              'sending the transaction'),
		('bal2',                 f'the {coin} balance'),

		('txcreate3',            'creating a transaction (spend from dev address to address :21)'),
		('txsign3',              'signing the transaction'),
		('txsend3',              'sending the transaction'),
		('bal3',                 f'the {coin} balance'),

		('tx_status1',           'getting the transaction status'),

		('txcreate4',            'creating a transaction (spend from MMGen address, low TX fee)'),
		('txbump',               'bumping the transaction fee'),

		('txsign4',              'signing the transaction'),
		('txsend4',              'sending the transaction'),
		('tx_status1a',          'getting the transaction status'),
		('bal4',                 f'the {coin} balance'),

		('txcreate5',            'creating a transaction (fund burn address)'),
		('txsign5',              'signing the transaction'),
		('txsend5',              'sending the transaction'),

		('addrimport_burn_addr', 'importing burn address'),
		('bal5',                 f'the {coin} balance'),

		('add_comment1',         'adding a UTF-8 label (zh)'),
		('chk_comment1',         'checking the label'),
		('add_comment2',         'adding a UTF-8 label (lat+cyr+gr)'),
		('chk_comment2',         'checking the label'),
		('remove_comment',       'removing the label'),
	),
	'contract': (
		'creating and deploying ERC20 tokens',
		('token_compile1',  'compiling ERC20 token #1'),
		('token_deploy1a',  'deploying ERC20 token #1 (SafeMath)'),
		('token_deploy1b',  'deploying ERC20 token #1 (Owned)'),
		('token_deploy1c',  'deploying ERC20 token #1 (Token)'),

		('tx_status2',      'getting the transaction status'),
		('bal6',            f'the {coin} balance'),

		('token_compile2',  'compiling ERC20 token #2'),
		('token_deploy2a',  'deploying ERC20 token #2 (SafeMath)'),
		('token_deploy2b',  'deploying ERC20 token #2 (Owned)'),
		('token_deploy2c',  'deploying ERC20 token #2 (Token)'),

		('contract_deploy', 'deploying contract (create,sign,send)'),
	),
	'token': (
		'creating, signing, sending and bumping ERC20 token transactions',

		('token_fund_users',           'transferring token funds from dev to user'),
		('token_user_bals',            'show balances after transfer'),
		('token_addrgen',              'generating token addresses'),
		('token_addrimport_badaddr1',  'importing token addresses (no token address)'),
		('token_addrimport_badaddr2',  'importing token addresses (bad token address)'),
		('token_addrimport_addr1',     'importing token addresses using token address (MM1)'),
		('token_addrimport_addr2',     'importing token addresses using token address (MM2)'),
		('token_addrimport_batch',     'importing token addresses (dummy batch mode) (MM1)'),
		('token_addrimport_sym',       'importing token addresses using token symbol (MM2)'),

		('bal7',                       f'the {coin} balance'),
		('token_bal1',                 f'the {coin} balance and token balance'),

		('token_txcreate1',            'creating a token transaction'),
		('token_txview1_raw',          'viewing the raw transaction'),
		('token_txsign1',              'signing the transaction'),
		('token_txsend1',              'sending the transaction'),
		('token_txview1_sig',          'viewing the signed transaction'),
		('tx_status3',                 'getting the transaction status'),
		('token_bal2',                 f'the {coin} balance and token balance'),

		('token_txcreate2',            'creating a token transaction (to burn address)'),
		('token_txbump',               'bumping the transaction fee'),

		('token_txsign2',              'signing the transaction'),
		('token_txsend2',              'sending the transaction'),
		('token_bal3',                 f'the {coin} balance and token balance'),

		('del_dev_addr',               'deleting the dev address'),

		('bal1_getbalance',            f'the {coin} balance (getbalance)'),

		('addrimport_token_burn_addr', 'importing the token burn address'),

		('token_bal4',                 f'the {coin} balance and token balance'),
		('token_bal_getbalance',       'the token balance (getbalance)'),

		('txcreate_noamt',             'creating a transaction (full amount send)'),
		('txsign_noamt',               'signing the transaction'),
		('txsend_noamt',               'sending the transaction'),

		('bal8',                       f'the {coin} balance'),
		('token_bal5',                 'the token balance'),

		('token_txcreate_noamt',       'creating a token transaction (full amount send)'),
		('token_txsign_noamt',         'signing the transaction'),
		('token_txsend_noamt',         'sending the transaction'),

		('bal9',                       f'the {coin} balance'),
		('token_bal6',                 'the token balance'),

		('listaddresses1',             'listaddresses'),
		('listaddresses2',             'listaddresses minconf=999999999 (ignored)'),
		('listaddresses3',             'listaddresses sort=age (ignored)'),
		('listaddresses4',             'listaddresses showempty=1 sort=age (ignored)'),

		('token_listaddresses1',       'listaddresses --token=mm1'),
		('token_listaddresses2',       'listaddresses --token=mm1 showempty=1'),
	),
	'twexport': (
		'exporting and importing tracking wallet to JSON',
		('twexport_noamt',       'exporting the tracking wallet (include_amts=0)'),
		('twmove',               'moving the tracking wallet'),
		('twimport',             'importing the tracking wallet'),
		('twview7',              'twview (cached_balances=1)'),
		('twview8',              'twview'),
		('twexport',             'exporting the tracking wallet'),
		('tw_chktotal',          'checking total value in tracking wallet dump'),
		('twmove',               'moving the tracking wallet'),
		('twimport',             'importing the tracking wallet'),
		('twcompare',            'comparing imported tracking wallet with original'),
		('edit_json_twdump',     'editing the tracking wallet JSON dump'),
		('twmove',               'moving the tracking wallet'),
		('twimport_nochksum',    'importing the edited tracking wallet JSON dump (ignore_checksum=1)'),

		('token_listaddresses3', 'listaddresses --token=mm1 showempty=1'),
		('token_listaddresses4', 'listaddresses --token=mm2 showempty=1'),
		('twview9',              'twview (check balance)'),
	),
	'cached': (
		'creating and sending transactions using cached balances',
		('twview_cached_balances',          'twview (cached balances)'),
		('token_twview_cached_balances',    'token twview (cached balances)'),
		('txcreate_cached_balances',        'txcreate (cached balances)'),
		('token_txcreate_cached_balances',  'token txcreate (cached balances)'),

		('txdo_cached_balances',            'txdo (cached balances)'),
		('txcreate_refresh_balances',       'refreshing balances'),
		('bal10',                           f'the {coin} balance'),

		('token_txdo_cached_balances',      'token txdo (cached balances)'),
		('token_txcreate_refresh_balances', 'refreshing token balances'),
		('token_bal7',                      'the token balance'),
	),
	'view': (
		'viewing addresses and unspent outputs',
		('twview1',       'twview'),
		('twview2',       'twview wide=1'),
		('twview3',       'twview wide=1 sort=age (ignored)'),
		('twview4',       'twview wide=1 minconf=999999999 (ignored)'),
		('twview5',       'twview wide=1 minconf=0 (ignored)'),
		('token_twview1', 'twview --token=mm1'),
		('token_twview2', 'twview --token=mm1 wide=1'),
		('token_twview3', 'twview --token=mm1 wide=1 sort=age (ignored)'),
	),
	'label': (
		'creating, editing and removing labels',
		('edit_comment1',       f'adding label to addr #{del_addrs[0]} in {coin} tracking wallet (zh)'),
		('edit_comment2',       f'editing label for addr #{del_addrs[0]} in {coin} tracking wallet (zh)'),
		('edit_comment3',       f'adding label to addr #{del_addrs[1]} in {coin} tracking wallet (lat+cyr+gr)'),
		('edit_comment4',       f'removing label from addr #{del_addrs[0]} in {coin} tracking wallet'),
		('token_edit_comment1', f'adding label to addr #{del_addrs[0]} in {coin} token tracking wallet'),
	),
	'remove': (
		'removing addresses from tracking wallet',
		('remove_addr1',       f'removing addr #{del_addrs[0]} from {coin} tracking wallet'),
		('twview6',            'twview (balance reduced after address removal)'),
		('remove_addr2',       f'removing addr #{del_addrs[1]} from {coin} tracking wallet'),
		('token_remove_addr1', f'removing addr #{del_addrs[0]} from {coin} token tracking wallet'),
		('token_remove_addr2', f'removing addr #{del_addrs[1]} from {coin} token tracking wallet'),
	),
	}

	def __init__(self,trunner,cfgs,spawn):
		CmdTestBase.__init__(self,trunner,cfgs,spawn)
		if trunner is None:
			return

		self.txcreate_args    = [f'--outdir={self.tmpdir}', '--quiet']
		self.eth_args         = [f'--outdir={self.tmpdir}', '--quiet']
		self.eth_args_noquiet = [f'--outdir={self.tmpdir}']

		from mmgen.protocol import init_proto
		self.proto = init_proto( cfg, cfg.coin, network='regtest', need_amt=True )

		from mmgen.daemon import CoinDaemon
		self.daemon = CoinDaemon( cfg, self.proto.coin+'_rt', test_suite=True )

		set_vbals(self.daemon.id)

		self.using_solc = check_solc_ver()
		if not self.using_solc:
			omsg(yellow('Using precompiled contract data'))

		omsg(blue(f'Coin daemon {self.daemon.id!r} selected'))

		self.genesis_fn = joinpath(self.tmpdir,'genesis.json')
		self.keystore_dir = os.path.relpath(joinpath(self.daemon.datadir,'keystore'))

		write_to_file(
			joinpath(self.tmpdir,parity_devkey_fn),
			dfl_devkey+'\n' )

		self.message = 'attack at dawn'
		self.spawn_env['MMGEN_BOGUS_SEND'] = ''

	@property
	async def rpc(self):
		from mmgen.rpc import rpc_init
		return await rpc_init(cfg,self.proto)

	async def setup(self):
		self.spawn('',msg_only=True)

		if not self.using_solc:
			srcdir = os.path.join(self.tr.repo_root,'test','ref','ethereum','bin')
			from shutil import copytree
			for d in ('mm1','mm2'):
				copytree(os.path.join(srcdir,d),os.path.join(self.tmpdir,d))

		d = self.daemon

		if d.id in ('geth','erigon'):
			self.genesis_setup(d)
			set_vt100()

		if d.id == 'erigon':
			self.write_to_tmpfile('signer_key',self.keystore_data['key']+'\n')
			d.usr_coind_args = [
				'--miner.sigfile={}'.format(os.path.join(self.tmpdir,'signer_key')),
				'--miner.etherbase={}'.format(self.keystore_data['address']) ]

		if d.id in ('geth','erigon'):
			imsg('  {:19} {}'.format('Cmdline:',' '.join(e for e in d.start_cmd if not 'verbosity' in e)))

		if not cfg.no_daemon_autostart:
			if not d.id in ('geth','erigon'):
				d.stop(silent=True)
				d.remove_datadir()
			d.start( silent = not (cfg.verbose or cfg.exact_output) )
			rpc = await self.rpc
			imsg(f'Daemon: {rpc.daemon.coind_name} v{rpc.daemon_version_str}')

		return 'ok'

	@property
	def keystore_data(self):
		if not hasattr(self,'_keystore_data'):

			wallet_fn = os.path.join( self.keystore_dir, os.listdir(self.keystore_dir)[0] )

			from mmgen.proto.eth.misc import decrypt_geth_keystore
			key = decrypt_geth_keystore(
				cfg       = cfg,
				wallet_fn = wallet_fn,
				passwd = b'' )

			with open(wallet_fn) as fh:
				res = json.loads(fh.read())

			res.update( { 'key': key.hex() } )
			self._keystore_data = res

		return self._keystore_data

	def genesis_setup(self,d):

		def make_key():
			pwfile = joinpath(self.tmpdir,'account_passwd')
			write_to_file(pwfile,'')
			run(['rm','-rf',self.keystore_dir])
			cmd = f'geth account new --password={pwfile} --lightkdf --keystore {self.keystore_dir}'
			cp = run(cmd.split(),stdout=PIPE,stderr=PIPE)
			if cp.returncode:
				die(1,cp.stderr.decode())

		def make_genesis(signer_addr,prealloc_addr):
			return {
				'config': {
					'chainId': 1337, # TODO: replace constant with var
					'homesteadBlock': 0,
					'eip150Block': 0,
					'eip155Block': 0,
					'eip158Block': 0,
					'byzantiumBlock': 0,
					'constantinopleBlock': 0,
					'petersburgBlock': 0,
					'istanbulBlock': 0,
					'muirGlacierBlock': 0,
					'berlinBlock': 0,
					'londonBlock': 0,
					'arrowGlacierBlock': 0,
					'grayGlacierBlock': 0,
					'shanghaiTime': 0,
					'terminalTotalDifficulty': 0,
					'terminalTotalDifficultyPassed': True,
					'isDev': True
				},
				'nonce': '0x0',
				'timestamp': '0x0',
				'extraData': '0x',
				'gasLimit': '0xaf79e0',
				'difficulty': '0x1',
				'mixHash': '0x0000000000000000000000000000000000000000000000000000000000000000',
				'coinbase': '0x0000000000000000000000000000000000000000',
				'number': '0x0',
				'gasUsed': '0x0',
				'parentHash': '0x0000000000000000000000000000000000000000000000000000000000000000',
				'baseFeePerGas': '0x3b9aca00',
				'excessBlobGas': None,
				'blobGasUsed': None,
				'alloc': {
					prealloc_addr: { 'balance': hex(prealloc_amt.toWei()) }
				}
			}

		def init_genesis(fn):
			cmd = f'{d.exec_fn} init --datadir {d.datadir} {fn}'
			cp = run( cmd.split(), stdout=PIPE, stderr=PIPE )
			if cp.returncode:
				die(1,cp.stderr.decode())

		d.stop(quiet=True)
		d.remove_datadir()

		imsg(cyan('Initializing Genesis Block:'))

		prealloc_amt = self.proto.coin_amt('1_000_000_000')

		make_key()
		signer_addr = self.keystore_data['address']
		self.write_to_tmpfile( 'signer_addr', signer_addr + '\n' )

		imsg(f'  Keystore:           {self.keystore_dir}')
		imsg(f'  Signer key:         {self.keystore_data["key"]}')
		imsg(f'  Signer address:     {signer_addr}')
		imsg(f'  Faucet:             {dfl_devaddr} ({prealloc_amt} ETH)')
		imsg(f'  Genesis block data: {self.genesis_fn}')

		genesis_data = make_genesis(signer_addr,dfl_devaddr)
		write_to_file( self.genesis_fn, json.dumps(genesis_data,indent='  ')+'\n' )
		init_genesis(self.genesis_fn)

	def daemon_version(self):
		t = self.spawn('mmgen-tool', self.eth_args + ['daemon_version'])
		t.expect('version')
		return t

	async def _wallet_upgrade(self, src_fn, expect1, expect2=None):
		if self.proto.coin == 'ETC':
			msg(f'skipping test {self.test_name!r} for ETC')
			return 'skip'
		from mmgen.tw.ctl import TwCtl
		twctl = await TwCtl(cfg, self.proto, no_wallet_init=True)
		from_fn = Path(ref_dir) / 'ethereum' / src_fn
		bak_fn = twctl.tw_dir / f'upgraded-{src_fn}'
		twctl.tw_dir.mkdir(mode=0o750, parents=True, exist_ok=True)
		dest = shutil.copy2(from_fn, twctl.tw_path)
		assert dest == twctl.tw_path, f'{dest} != {twctl.tw_path}'
		t = self.spawn('mmgen-tool', self.eth_args + ['twview'])
		t.expect(expect1)
		if expect2:
			t.expect(expect2)
		t.read()
		twctl.tw_path.rename(bak_fn)
		return t

	async def wallet_upgrade1(self):
		return await self._wallet_upgrade('tracking-wallet-v1.json', 'accounts field', 'network field')

	async def wallet_upgrade2(self):
		return await self._wallet_upgrade('tracking-wallet-v2.json', 'token params field', 'network field')

	def addrgen(self,addrs='1-3,11-13,21-23'):
		t = self.spawn('mmgen-addrgen', self.eth_args + [dfl_words_file,addrs])
		t.written_to_file('Addresses')
		return t

	def addrimport(
			self,
			ext       = '21-23]{}.regtest.addrs',
			expect    = '9/9',
			add_args  = [],
			bad_input = False,
			exit_val  = None):
		ext = ext.format('-α' if cfg.debug_utf8 else '')
		fn = self.get_file_with_ext(ext,no_dot=True,delete=False)
		t = self.spawn('mmgen-addrimport', self.eth_args[1:-1] + add_args + [fn], exit_val=exit_val)
		if bad_input:
			return t
		t.expect('Importing')
		t.expect(expect)
		return t

	def addrimport_one_addr(self,addr=None,extra_args=[]):
		t = self.spawn('mmgen-addrimport', self.eth_args[1:] + extra_args + ['--address='+addr])
		t.expect('OK')
		return t

	def addrimport_dev_addr(self):
		return self.addrimport_one_addr(addr=dfl_devaddr)

	def addrimport_burn_addr(self):
		return self.addrimport_one_addr(addr=burn_addr)

	def txcreate(
			self,
			args            = [],
			menu            = [],
			acct            = '1',
			caller          = 'txcreate',
			interactive_fee = '50G',
			fee_info_data   = ('0.00105','50'),
			no_read         = False,
			print_listing   = True,
			tweaks          = []):
		fee_info_pat = r'\D{}\D.*{c} .*\D{}\D.*gas price in Gwei'.format( *fee_info_data, c=self.proto.coin )
		t = self.spawn('mmgen-'+caller, self.txcreate_args + ['-B'] + args)
		if print_listing:
			t.expect(r'add \[l\]abel, .*?:.','p', regex=True)
			t.written_to_file('Account balances listing')
		t = self.txcreate_ui_common(
				t,
				menu              = menu,
				caller            = caller,
				input_sels_prompt = 'to spend from',
				inputs            = acct,
				file_desc         = 'transaction',
				bad_input_sels    = True,
				interactive_fee   = interactive_fee,
				fee_info_pat      = fee_info_pat,
				fee_desc          = 'transaction fee or gas price',
				add_comment       = tx_comment_jp,
				tweaks            = tweaks)
		if not no_read:
			t.read()
		return t

	def txsign(self,ni=False,ext='{}.regtest.rawtx',add_args=[],dev_send=False):
		ext = ext.format('-α' if cfg.debug_utf8 else '')
		keyfile = joinpath(self.tmpdir,parity_devkey_fn)
		txfile = self.get_file_with_ext(ext,no_dot=True)
		t = self.spawn(
				'mmgen-txsign',
				self.eth_args
				+ [f'--coin={self.proto.coin}']
				+ ['--rpc-host=bad_host'] # ETH signing must work without RPC
				+ add_args
				+ ([],['--yes'])[ni]
				+ ([f'--keys-from-file={keyfile}'] if dev_send else [])
				+ [txfile, dfl_words_file])
		return self.txsign_ui_common(t,ni=ni,has_label=True)

	def txsend(self,ext='{}.regtest.sigtx',add_args=[]):
		ext = ext.format('-α' if cfg.debug_utf8 else '')
		txfile = self.get_file_with_ext(ext,no_dot=True)
		t = self.spawn('mmgen-txsend', self.eth_args + add_args + [txfile])
		self.txsend_ui_common(
				t,
				quiet      = not cfg.debug,
				bogus_send = False,
				has_label  = True)
		return t

	def txview(self,ext_fs):
		ext = ext_fs.format('-α' if cfg.debug_utf8 else '')
		txfile = self.get_file_with_ext(ext,no_dot=True)
		return self.spawn( 'mmgen-tool',['--verbose','txview',txfile] )

	def fund_dev_address(self):
		"""
		For Erigon, fund the default (Parity) dev address from the Erigon dev address
		For the others, send a junk TX to keep block counts equal for all daemons
		"""
		dt = namedtuple('data',['devkey_fn','dest','amt'])
		d = dt( parity_devkey_fn, burn_addr2, '1' )
		t = self.txcreate(
			args    = self.eth_args_noquiet + [
				f'--keys-from-file={joinpath(self.tmpdir,d.devkey_fn)}',
				f'{d.dest},{d.amt}',
			],
			menu    = ['a','r'],
			caller  = 'txdo',
			acct    = '1',
			no_read = True )
		self._do_confirm_send(t,quiet=not cfg.debug,sure=False)
		t.read()
		self.get_file_with_ext('sigtx',delete_all=True)
		return t

	def txcreate1(self):
		# include one invalid keypress 'X' -- see EthereumTwUnspentOutputs.key_mappings
		menu = ['a','d','r','M','X','e','m','m']
		args = ['98831F3A:E:1,123.456']
		return self.txcreate(args=args,menu=menu,acct='1',tweaks=['confirm_non_mmgen'])
	def txview1_raw(self):
		return self.txview(ext_fs='{}.regtest.rawtx')
	def txsign1(self):
		return self.txsign(add_args=['--use-internal-keccak-module'],dev_send=True)
	def tx_status0_bad(self):
		return self.tx_status(ext='{}.regtest.sigtx',expect_str='neither in mempool nor blockchain',exit_val=1)
	def txsign1_ni(self):
		return self.txsign(ni=True,dev_send=True)
	def txsend1(self):
		return self.txsend()
	def txview1_sig(self): # do after send so that TxID is displayed
		return self.txview(ext_fs='{}.regtest.sigtx')
	def bal1(self):
		return self.bal(n='1')

	def txcreate2(self):
		args = ['98831F3A:E:11,1.234']
		return self.txcreate(args=args,acct='10',tweaks=['confirm_non_mmgen'])
	def txsign2(self):
		return self.txsign(ni=True,ext='1.234,50000]{}.regtest.rawtx',dev_send=True)
	def txsend2(self):
		return self.txsend(ext='1.234,50000]{}.regtest.sigtx')
	def bal2(self):
		return self.bal(n='2')

	def txcreate3(self):
		args = ['98831F3A:E:21,2.345']
		return self.txcreate(args=args,acct='10',tweaks=['confirm_non_mmgen'])
	def txsign3(self):
		return self.txsign(ni=True,ext='2.345,50000]{}.regtest.rawtx',dev_send=True)
	def txsend3(self):
		return self.txsend(ext='2.345,50000]{}.regtest.sigtx')
	def bal3(self):
		return self.bal(n='3')

	def tx_status(self,ext,expect_str,expect_str2='',add_args=[],exit_val=0):
		ext = ext.format('-α' if cfg.debug_utf8 else '')
		txfile = self.get_file_with_ext(ext,no_dot=True)
		t = self.spawn(
				'mmgen-txsend',
				self.eth_args + add_args + ['--status', txfile],
				exit_val = exit_val)
		t.expect(expect_str)
		if expect_str2:
			t.expect(expect_str2)
		return t

	def tx_status1(self):
		return self.tx_status(ext='2.345,50000]{}.regtest.sigtx',expect_str='has 1 confirmation')

	def tx_status1a(self):
		return self.tx_status(ext='2.345,50000]{}.regtest.sigtx',expect_str='has 2 confirmations')

	async def msgsign_chk(self): # NB: Geth only!

		def create_signature_mmgen():
			key = self.keystore_data['key']
			imsg(f'Key:       {key}')
			from mmgen.proto.eth.misc import ec_sign_message_with_privkey
			return ec_sign_message_with_privkey(cfg,self.message,bytes.fromhex(key),'eth_sign')

		async def create_signature_rpc():
			addr = self.read_from_tmpfile('signer_addr').strip()
			imsg(f'Address:   {addr}')
			rpc = await self.rpc
			return await rpc.call(
				'eth_sign',
				'0x' + addr,
				'0x' + self.message.encode().hex() )

		if not self.daemon.id == 'geth':
			return 'skip'

		self.spawn('',msg_only=True)

		sig = '0x' + create_signature_mmgen()
		sig_chk = await create_signature_rpc()

		# Compare signatures
		imsg(f'Message:   {self.message}')
		imsg(f'Signature: {sig}')
		cmp_or_die(sig,sig_chk,'message signatures')
		imsg('Geth and MMGen signatures match')

		return 'ok'

	def msgcreate(self,add_args=[]):
		t = self.spawn('mmgen-msg', self.eth_args_noquiet + add_args + [ 'create', self.message, '98831F3A:E:1' ])
		t.written_to_file('Unsigned message data')
		return t

	def msgsign(self):
		fn = get_file_with_ext(self.tmpdir,'rawmsg.json')
		t = self.spawn('mmgen-msg', self.eth_args_noquiet + [ 'sign', fn, dfl_words_file ])
		t.written_to_file('Signed message data')
		return t

	def msgverify(self,fn=None,msghash_type='eth_sign'):
		fn = fn or get_file_with_ext(self.tmpdir,'sigmsg.json')
		t = self.spawn('mmgen-msg', self.eth_args_noquiet + [ 'verify', fn ])
		t.expect(msghash_type)
		t.expect('1 signature verified')
		return t

	def msgexport(self):
		fn = get_file_with_ext(self.tmpdir,'sigmsg.json')
		t = self.spawn('mmgen-msg', self.eth_args_noquiet + [ 'export', fn ])
		t.written_to_file('Signature data')
		return t

	def msgverify_export(self):
		return self.msgverify(
			fn = os.path.join(self.tmpdir,'signatures.json') )

	def msgcreate_raw(self):
		get_file_with_ext(self.tmpdir,'rawmsg.json',delete_all=True)
		return self.msgcreate(add_args=['--msghash-type=raw'])

	def msgsign_raw(self):
		get_file_with_ext(self.tmpdir,'sigmsg.json',delete_all=True)
		return self.msgsign()

	def msgverify_raw(self):
		return self.msgverify(msghash_type='raw')

	def msgexport_raw(self):
		get_file_with_ext(self.tmpdir,'signatures.json',no_dot=True,delete_all=True)
		return self.msgexport()

	def msgverify_export_raw(self):
		return self.msgverify(
			fn = os.path.join(self.tmpdir,'signatures.json'),
			msghash_type = 'raw' )

	def txcreate4(self):
		return self.txcreate(
			args             = ['98831F3A:E:2,23.45495'],
			acct             = '1',
			interactive_fee  = '40G',
			fee_info_data    = ('0.00084','40') )

	def txbump(self,ext=',40000]{}.regtest.rawtx',fee='50G',add_args=[]):
		ext = ext.format('-α' if cfg.debug_utf8 else '')
		txfile = self.get_file_with_ext(ext,no_dot=True)
		t = self.spawn('mmgen-txbump', self.eth_args + add_args + ['--yes',txfile])
		t.expect('or gas price: ',fee+'\n')
		return t

	def txsign4(self):
		return self.txsign(ni=True,ext='.45495,50000]{}.regtest.rawtx',dev_send=True)
	def txsend4(self):
		return self.txsend(ext='.45495,50000]{}.regtest.sigtx')
	def bal4(self):
		return self.bal(n='4')

	def txcreate5(self):
		args = [burn_addr + ','+amt1]
		return self.txcreate(args=args,acct='10',tweaks=['confirm_non_mmgen'])
	def txsign5(self):
		return self.txsign(ni=True,ext=amt1+',50000]{}.regtest.rawtx',dev_send=True)
	def txsend5(self):
		return self.txsend(ext=amt1+',50000]{}.regtest.sigtx')
	def bal5(self):
		return self.bal(n='5')

	#bal_corr = Decimal('0.0000032') # gas use for token sends varies between ETH and ETC!
	bal_corr = Decimal('0.0000000') # update: OpenEthereum team seems to have corrected this

	def bal(self,n):
		t = self.spawn('mmgen-tool', self.eth_args + ['twview','wide=1'])
		text = t.read(strip_color=True)
		for b in bals(n):
			addr,amt,adj = b if len(b) == 3 else b + (False,)
			if adj and self.proto.coin == 'ETC':
				amt = str(Decimal(amt) + Decimal(adj[1]) * self.bal_corr)
			pat = r'\D{}\D.*\D{}\D'.format( addr, amt.replace('.',r'\.') )
			assert re.search(pat,text), pat
		ss = f'Total {self.proto.coin}:'
		assert re.search(ss,text),ss
		return t

	def token_bal(self,n=None):
		t = self.spawn('mmgen-tool', self.eth_args + ['--token=mm1','twview','wide=1'])
		text = t.read(strip_color=True)
		for b in token_bals(n):
			addr,_amt1,_amt2,adj = b if len(b) == 4 else b + (False,)
			if adj and self.proto.coin == 'ETC':
				_amt2 = str(Decimal(_amt2) + Decimal(adj[1]) * self.bal_corr)
			pat = fr'{addr}\b.*\D{_amt1}\D.*\b{_amt2}\D'
			assert re.search(pat,text), pat
		ss = 'Total MM1:'
		assert re.search(ss,text),ss
		return t

	def bal_getbalance(self,sid,idx,etc_adj=False,extra_args=[]):
		bal1 = token_bals_getbalance(idx)[0]
		bal2 = token_bals_getbalance(idx)[1]
		bal1 = Decimal(bal1)
		if etc_adj and self.proto.coin == 'ETC':
			bal1 += self.bal_corr
		t = self.spawn('mmgen-tool', self.eth_args + extra_args + ['getbalance'])
		t.expect(rf'{sid}:.*'+str(bal1),regex=True)
		t.expect(r'Non-MMGen:.*'+bal2,regex=True)
		total = strip_ansi_escapes(t.expect_getend(rf'TOTAL {self.proto.coin}')).split()[0]
		assert Decimal(bal1) + Decimal(bal2) == Decimal(total)
		return t

	def add_comment(self,comment,addr='98831F3A:E:3'):
		t = self.spawn('mmgen-tool', self.eth_args + ['add_label',addr,comment])
		t.expect('Added label.*in tracking wallet',regex=True)
		return t

	def chk_comment(self,comment_pat,addr='98831F3A:E:3'):
		t = self.spawn('mmgen-tool', self.eth_args + ['listaddresses','all_labels=1'])
		t.expect(fr'{addr}\b.*{comment_pat}',regex=True)
		return t

	def add_comment1(self):
		return self.add_comment(comment=tw_comment_zh)
	def chk_comment1(self):
		return self.chk_comment(comment_pat=tw_comment_zh[:3])
	def add_comment2(self):
		return self.add_comment(comment=tw_comment_lat_cyr_gr)
	def chk_comment2(self):
		return self.chk_comment(comment_pat=tw_comment_lat_cyr_gr[:3])

	def remove_comment(self,addr='98831F3A:E:3'):
		t = self.spawn('mmgen-tool', self.eth_args + ['remove_label',addr])
		t.expect('Removed label.*in tracking wallet',regex=True)
		return t

	def token_compile(self,token_data={}):
		odir = joinpath(self.tmpdir,token_data['symbol'].lower())
		if not self.using_solc:
			imsg(f'Using precompiled contract data in {odir}')
			return 'skip' if os.path.exists(odir) else False
		self.spawn('',msg_only=True)
		cmd_args = [f'--{k}={v}' for k,v in list(token_data.items())]
		imsg("Compiling solidity token contract '{}' with 'solc'".format( token_data['symbol'] ))
		try:
			os.mkdir(odir)
		except:
			pass
		cmd = [
			'python3',
			'scripts/create-token.py',
			'--coin=' + self.proto.coin,
			'--outdir=' + odir
		] + cmd_args + [self.proto.checksummed_addr(dfl_devaddr)]
		imsg('Executing: {}'.format( ' '.join(cmd) ))
		cp = run(cmd,stdout=DEVNULL,stderr=PIPE)
		if cp.returncode != 0:
			rmsg('solc failed with the following output:')
			die(2,cp.stderr.decode())
		imsg('ERC20 token {!r} compiled'.format( token_data['symbol'] ))
		return 'ok'

	def token_compile1(self):
		token_data = { 'name':'MMGen Token 1', 'symbol':'MM1', 'supply':10**26, 'decimals':18 }
		return self.token_compile(token_data)

	def token_compile2(self):
		token_data = { 'name':'MMGen Token 2', 'symbol':'MM2', 'supply':10**18, 'decimals':10 }
		return self.token_compile(token_data)

	async def get_tx_receipt(self,txid):
		if self.daemon.id == 'geth': # yet another Geth bug
			await asyncio.sleep(0.5)
		from mmgen.tx import NewTX
		tx = await NewTX(cfg=cfg,proto=self.proto)
		tx.rpc = await self.rpc
		res = await tx.get_receipt(txid)
		imsg(f'Gas sent:  {res.gas_sent.hl():<9} {(res.gas_sent*res.gas_price).hl2(encl="()")}')
		imsg(f'Gas used:  {res.gas_used.hl():<9} {(res.gas_used*res.gas_price).hl2(encl="()")}')
		imsg(f'Gas price: {res.gas_price.hl()}')
		if res.gas_used == res.gas_sent:
			omsg(yellow('Warning: all gas was used!'))
		return res

	async def token_deploy(self,num,key,gas,mmgen_cmd='txdo',tx_fee='8G'):
		keyfile = joinpath(self.tmpdir,parity_devkey_fn)
		fn = joinpath(self.tmpdir,'mm'+str(num),key+'.bin')
		args = [
			'-B',
			f'--fee={tx_fee}',
			f'--gas={gas}',
			f'--contract-data={fn}',
			f'--inputs={dfl_devaddr}',
			'--yes',
		]
		if mmgen_cmd == 'txdo':
			args += ['-k',keyfile]
		t = self.spawn( 'mmgen-'+mmgen_cmd, self.eth_args + args)
		if mmgen_cmd == 'txcreate':
			t.written_to_file('transaction')
			ext = '[0,8000]{}.regtest.rawtx'.format('-α' if cfg.debug_utf8 else '')
			txfile = self.get_file_with_ext(ext,no_dot=True)
			t = self.spawn('mmgen-txsign', self.eth_args + ['--yes','-k',keyfile,txfile],no_msg=True)
			self.txsign_ui_common(t,ni=True)
			txfile = txfile.replace('.rawtx','.sigtx')
			t = self.spawn('mmgen-txsend', self.eth_args + [txfile],no_msg=True)

		txid = self.txsend_ui_common(t,
			caller = mmgen_cmd,
			quiet  = mmgen_cmd == 'txdo' or not cfg.debug,
			bogus_send = False )
		addr = strip_ansi_escapes(t.expect_getend('Contract address: '))
		if (await self.get_tx_receipt(txid)).status == 0:
			die(2,f'Contract {num}:{key} failed to execute. Aborting')
		if key == 'Token':
			self.write_to_tmpfile( f'token_addr{num}', addr+'\n' )
			imsg(f'\nToken MM{num} deployed!')
		return t

	async def token_deploy1a(self):
		return await self.token_deploy(num=1,key='SafeMath',gas=500_000)
	async def token_deploy1b(self):
		return await self.token_deploy(num=1,key='Owned',   gas=1_000_000)
	async def token_deploy1c(self):
		return await self.token_deploy(num=1,key='Token',   gas=4_000_000,tx_fee='7G')

	def tx_status2(self):
		return self.tx_status(
				ext        = self.proto.coin+'[0,7000]{}.regtest.sigtx',
				expect_str = 'successfully executed')

	def bal6(self):
		return self.bal5()

	async def token_deploy2a(self):
		return await self.token_deploy(num=2,key='SafeMath',gas=500_000)
	async def token_deploy2b(self):
		return await self.token_deploy(num=2,key='Owned',   gas=1_000_000)
	async def token_deploy2c(self):
		return await self.token_deploy(num=2,key='Token',   gas=4_000_000)

	async def contract_deploy(self): # test create,sign,send
		return await self.token_deploy(num=2,key='SafeMath',gas=500_000,mmgen_cmd='txcreate')

	async def token_transfer_ops(self,op,amt=1000,num_tokens=2):
		self.spawn('',msg_only=True)
		sid = dfl_sid
		from mmgen.tool.wallet import tool_cmd
		usr_mmaddrs = [f'{sid}:E:{i}' for i in (11,21)][:num_tokens]

		from mmgen.proto.eth.contract import ResolvedToken
		async def do_transfer(rpc):
			for i in range(num_tokens):
				tk = await ResolvedToken(
					cfg,
					self.proto,
					rpc,
					self.read_from_tmpfile(f'token_addr{i+1}').strip() )
				imsg_r( '\n' + await tk.info() )
				imsg('dev token balance (pre-send): {}'.format( await tk.get_balance(dfl_devaddr) ))
				imsg(f'Sending {amt} {self.proto.dcoin} to address {usr_addrs[i]} ({usr_mmaddrs[i]})')
				txid = await tk.transfer(
					dfl_devaddr,
					usr_addrs[i],
					amt,
					dfl_devkey,
					start_gas = self.proto.coin_amt(60000,'wei'),
					gasPrice  = self.proto.coin_amt(8,'Gwei') )
				if (await self.get_tx_receipt(txid)).status == 0:
					die(2,'Transfer of token funds failed. Aborting')

		async def show_bals(rpc):
			for i in range(num_tokens):
				tk = await ResolvedToken(
					cfg,
					self.proto,
					rpc,
					self.read_from_tmpfile(f'token_addr{i+1}').strip() )
				imsg('Token: {}'.format( await tk.get_symbol() ))
				imsg(f'dev token balance: {await tk.get_balance(dfl_devaddr)}')
				imsg('usr token balance: {} ({} {})'.format(
					await tk.get_balance(usr_addrs[i]),
					usr_mmaddrs[i],
					usr_addrs[i] ))

		def gen_addr(addr):
			return tool_cmd(cfg,cmdname='gen_addr',proto=self.proto).gen_addr(addr,dfl_words_file)

		silence()
		usr_addrs = list(map(gen_addr,usr_mmaddrs))
		if op == 'show_bals':
			await show_bals(await self.rpc)
		elif op == 'do_transfer':
			await do_transfer(await self.rpc)
		end_silence()
		return 'ok'

	def token_fund_users(self):
		return self.token_transfer_ops(op='do_transfer')

	def token_user_bals(self):
		return self.token_transfer_ops(op='show_bals')

	def token_addrgen(self,num_tokens=2):
		t = self.addrgen(addrs='11-13')
		if num_tokens == 1:
			return t
		ok_msg()
		return self.addrgen(addrs='21-23')

	def token_addrimport_badaddr1(self):
		t = self.addrimport(
			ext       = '[11-13]{}.regtest.addrs',
			add_args  = ['--token=abc'],
			bad_input = True,
			exit_val  = 2)
		t.expect('could not be resolved')
		return t

	def token_addrimport_badaddr2(self):
		t = self.addrimport(
			ext       = '[11-13]{}.regtest.addrs',
			add_args  = ['--token='+'00deadbeef'*4],
			bad_input = True,
			exit_val  = 2)
		t.expect('could not be resolved')
		return t

	def token_addrimport(self,addr_file,addr_range,expect,extra_args=[]):
		token_addr = self.read_from_tmpfile(addr_file).strip()
		return self.addrimport(
			ext      = f'[{addr_range}]{{}}.regtest.addrs',
			expect   = expect,
			add_args = ['--token-addr='+token_addr]+extra_args )

	def token_addrimport_addr1(self):
		return self.token_addrimport('token_addr1','11-13',expect='3/3')

	def token_addrimport_addr2(self):
		return self.token_addrimport('token_addr2','21-23',expect='3/3')

	def token_addrimport_batch(self):
		return self.token_addrimport('token_addr1','11-13',expect='3 addresses',extra_args=['--batch'])

	def token_addrimport_sym(self):
		return self.addrimport(
			ext      = '[21-23]{}.regtest.addrs',
			expect   = '3/3',
			add_args = ['--token=MM2'] )

	def bal7(self):
		return self.bal5()
	def token_bal1(self):
		return self.token_bal(n='1')

	def token_txcreate(self,args=[],token='',inputs='1',fee='50G',file_desc='Unsigned transaction'):
		return self.txcreate_ui_common(
			self.spawn('mmgen-txcreate', self.txcreate_args + ['--token='+token,'-B','--fee='+fee] + args),
			menu              = [],
			inputs            = inputs,
			input_sels_prompt = 'to spend from',
			add_comment       = tx_comment_lat_cyr_gr,
			file_desc         = file_desc)
	def token_txsign(self,ext='',token=''):
		return self.txsign(ni=True,ext=ext,add_args=['--token='+token])
	def token_txsend(self,ext='',token=''):
		return self.txsend(ext=ext,add_args=['--token='+token])

	def token_txcreate1(self):
		return self.token_txcreate(args=['98831F3A:E:12,1.23456'],token='mm1')
	def token_txview1_raw(self):
		return self.txview(ext_fs='1.23456,50000]{}.regtest.rawtx')
	def token_txsign1(self):
		return self.token_txsign(ext='1.23456,50000]{}.regtest.rawtx',token='mm1')
	def token_txsend1(self):
		return self.token_txsend(ext='1.23456,50000]{}.regtest.sigtx',token='mm1')
	def token_txview1_sig(self):
		return self.txview(ext_fs='1.23456,50000]{}.regtest.sigtx')

	def tx_status3(self):
		return self.tx_status(
			ext         = '1.23456,50000]{}.regtest.sigtx',
			add_args    = ['--token=mm1'],
			expect_str  = 'successfully executed',
			expect_str2 = 'has 1 confirmation' )

	def token_bal2(self):
		return self.token_bal(n='2')

	def twview(self,args=[],expect_str='',tool_args=[]):
		t = self.spawn('mmgen-tool', self.eth_args + args + ['twview'] + tool_args)
		if expect_str:
			t.expect(expect_str,regex=True)
		return t

	def token_txcreate2(self):
		return self.token_txcreate(args=[burn_addr+','+amt2],token='mm1')
	def token_txbump(self):
		return self.txbump(ext=amt2+',50000]{}.regtest.rawtx',fee='56G',add_args=['--token=mm1'])
	def token_txsign2(self):
		return self.token_txsign(ext=amt2+',50000]{}.regtest.rawtx',token='mm1')
	def token_txsend2(self):
		return self.token_txsend(ext=amt2+',50000]{}.regtest.sigtx',token='mm1')

	def token_bal3(self):
		return self.token_bal(n='3')

	def del_dev_addr(self):
		t = self.spawn('mmgen-tool', self.eth_args + ['remove_address', dfl_devaddr])
		t.expect(f"'{dfl_devaddr}' deleted")
		return t

	def bal1_getbalance(self):
		return self.bal_getbalance(dfl_sid,'1',etc_adj=True)

	def addrimport_token_burn_addr(self):
		return self.addrimport_one_addr(addr=burn_addr,extra_args=['--token=mm1'])

	def token_bal4(self):
		return self.token_bal(n='4')

	def token_bal_getbalance(self):
		return self.bal_getbalance(dfl_sid,'2',extra_args=['--token=mm1'])

	def txcreate_noamt(self):
		return self.txcreate(args=['98831F3A:E:12'])
	def txsign_noamt(self):
		return self.txsign(ext='99.99895,50000]{}.regtest.rawtx')
	def txsend_noamt(self):
		return self.txsend(ext='99.99895,50000]{}.regtest.sigtx')

	def bal8(self):
		return self.bal(n='8')
	def token_bal5(self):
		return self.token_bal(n='5')

	def token_txcreate_noamt(self):
		return self.token_txcreate(args=['98831F3A:E:13'],token='mm1',inputs='2',fee='51G')
	def token_txsign_noamt(self):
		return self.token_txsign(ext='1.23456,51000]{}.regtest.rawtx',token='mm1')
	def token_txsend_noamt(self):
		return self.token_txsend(ext='1.23456,51000]{}.regtest.sigtx',token='mm1')

	def bal9(self):
		return self.bal(n='9')
	def token_bal6(self):
		return self.token_bal(n='6')

	def listaddresses(self,args=[],tool_args=['all_labels=1']):
		return self.spawn('mmgen-tool', self.eth_args + args + ['listaddresses'] + tool_args)

	def listaddresses1(self):
		return self.listaddresses()
	def listaddresses2(self):
		return self.listaddresses(tool_args=['minconf=999999999'])
	def listaddresses3(self):
		return self.listaddresses(tool_args=['sort=amt','reverse=1'])
	def listaddresses4(self):
		return self.listaddresses(tool_args=['sort=age','showempty=0'])

	def token_listaddresses1(self):
		return self.listaddresses(args=['--token=mm1'])
	def token_listaddresses2(self):
		return self.listaddresses(args=['--token=mm1'],tool_args=['showempty=1'])
	def token_listaddresses3(self):
		return self.listaddresses(args=['--token=mm1'],tool_args=['showempty=0'])
	def token_listaddresses4(self):
		return self.listaddresses(args=['--token=mm2'],tool_args=['sort=age','reverse=1'])

	def twview_cached_balances(self):
		return self.twview(args=['--cached-balances'])
	def token_twview_cached_balances(self):
		return self.twview(args=['--token=mm1','--cached-balances'])

	def txcreate_cached_balances(self):
		args = ['--fee=20G','--cached-balances','98831F3A:E:3,0.1276']
		return self.txcreate(args=args,acct='2')
	def token_txcreate_cached_balances(self):
		args=['--cached-balances','--fee=12G','98831F3A:E:12,1.2789']
		return self.token_txcreate(args=args,token='mm1')

	def txdo_cached_balances(
			self,
			acct          = '2',
			fee_info_data = ('0.00105','50'),
			add_args      = ['98831F3A:E:3,0.4321']):
		t = self.txcreate(
				args          = ['--fee=20G','--cached-balances'] + add_args + [dfl_words_file],
				acct          = acct,
				caller        = 'txdo',
				fee_info_data = fee_info_data,
				no_read       = True)
		self._do_confirm_send(t,quiet=not cfg.debug,sure=False)
		return t

	def txcreate_refresh_balances(self):
		return self._txcreate_refresh_balances(
			bals       = ['2','3'],
			args       = ['-B','--cached-balances','-i'],
			total      = vbal5,
			adj_total  = True,
			total_coin = None)

	def _txcreate_refresh_balances(self,bals,args,total,adj_total,total_coin):

		if total_coin is None:
			total_coin = self.proto.coin

		if self.proto.coin == 'ETC' and adj_total:
			total = str(Decimal(total) + self.bal_corr)
		t = self.spawn('mmgen-txcreate', self.txcreate_args + args)
		for n in bals:
			t.expect('[R]efresh balance:\b','R')
			t.expect(' main menu): ',n+'\n')
			t.expect('Is this what you want? (y/N): ','y')
		t.expect('[R]efresh balance:\b','q')
		t.expect(rf'Total unspent:.*\D{total}\D.*{total_coin}',regex=True)
		return t

	def bal10(self):
		return self.bal(n='10')

	def token_txdo_cached_balances(self):
		return self.txdo_cached_balances(
			acct          = '1',
			fee_info_data = ('0.0026','50'),
			add_args      = ['--token=mm1','98831F3A:E:12,43.21'] )

	def token_txcreate_refresh_balances(self):
		return self._txcreate_refresh_balances(
			bals       = ['1','2'],
			args       = ['--token=mm1','-B','--cached-balances','-i'],
			total      = '1000',
			adj_total  = False,
			total_coin = 'MM1' )

	def token_bal7(self):
		return self.token_bal(n='7')

	def twview1(self):
		return self.twview()
	def twview2(self):
		return self.twview(tool_args=['wide=1'])
	def twview3(self):
		return self.twview(tool_args=['wide=1','sort=age'])
	def twview4(self):
		return self.twview(tool_args=['wide=1','minconf=999999999'])
	def twview5(self):
		return self.twview(tool_args=['wide=1','minconf=0'])
	def twview6(self):
		return self.twview(expect_str=vbal7)
	def twview7(self):
		return self.twview(args=['--cached-balances'])
	def twview8(self):
		return self.twview()
	def twview9(self):
		return self.twview(args=['--cached-balances'],expect_str=vbal6)

	def token_twview1(self):
		return self.twview(args=['--token=mm1'])
	def token_twview2(self):
		return self.twview(args=['--token=mm1'],tool_args=['wide=1'])
	def token_twview3(self):
		return self.twview(args=['--token=mm1'],tool_args=['wide=1','sort=age'])

	def edit_comment(
			self,
			out_num,
			args          = [],
			action        = 'l',
			comment_text  = None,
			changed       = False,
			pexpect_spawn = None):
		t = self.spawn('mmgen-txcreate', self.txcreate_args + args + ['-B','-i'],pexpect_spawn=pexpect_spawn)

		menu_prompt = 'efresh balance:\b'

		t.expect(menu_prompt,'M')
		t.expect(menu_prompt,action)
		t.expect(r'return to main menu): ',out_num+'\n')

		for p,r in (
			('Enter label text.*: ',comment_text+'\n') if comment_text is not None else (r'\(y/N\): ','y'),
			(r'\(y/N\): ','y') if comment_text == Ctrl_U else (None,None),
		):
			if p:
				t.expect(p,r,regex=True)

		m = (
			'Label for account #{} edited' if changed else
			'Account #{} removed' if action == 'D' else
			'Label added to account #{}' if comment_text and comment_text != Ctrl_U else
			'Label removed from account #{}' )

		t.expect(m.format(out_num))
		t.expect(menu_prompt,'M')
		t.expect(menu_prompt,'q')

		t.expect('Total unspent:')

		return t

	def edit_comment1(self):
		return self.edit_comment(out_num=del_addrs[0],comment_text=tw_comment_zh[:3])
	def edit_comment2(self):
		spawn = not sys.platform == 'win32'
		return self.edit_comment(
				out_num       = del_addrs[0],
				comment_text  = tw_comment_zh[3:],
				changed       = True,
				pexpect_spawn = spawn)
	def edit_comment3(self):
		return self.edit_comment(out_num=del_addrs[1],comment_text=tw_comment_lat_cyr_gr)
	def edit_comment4(self):
		if self.skip_for_win():
			return 'skip'
		return self.edit_comment(out_num=del_addrs[0],comment_text=Ctrl_U,pexpect_spawn=True)

	def token_edit_comment1(self):
		return self.edit_comment(out_num='1',comment_text='Token label #1',args=['--token=mm1'])

	def remove_addr1(self):
		return self.edit_comment(out_num=del_addrs[0],action='D')
	def remove_addr2(self):
		return self.edit_comment(out_num=del_addrs[1],action='D')
	def token_remove_addr1(self):
		return self.edit_comment(out_num=del_addrs[0],args=['--token=mm1'],action='D')
	def token_remove_addr2(self):
		return self.edit_comment(out_num=del_addrs[1],args=['--token=mm1'],action='D')

	def twexport_noamt(self):
		return self.twexport(add_args=['include_amts=0'])

	def twexport(self,add_args=[]):
		t = self.spawn('mmgen-tool', self.eth_args + ['twexport'] + add_args)
		t.written_to_file('JSON data')
		return t

	async def twmove(self):
		self.spawn('',msg_only=True)
		from mmgen.tw.ctl import TwCtl
		twctl = await TwCtl(cfg, self.proto, no_wallet_init=True)
		imsg('Moving tracking wallet')
		fn_bak = twctl.tw_path.with_suffix('.bak.json')
		fn_bak.unlink(missing_ok=True)
		twctl.tw_path.rename(fn_bak)
		return 'ok'

	def twimport(self,add_args=[],expect_str=None):
		from mmgen.tw.json import TwJSON
		fn = joinpath( self.tmpdir, TwJSON.Base(cfg,self.proto).dump_fn )
		t = self.spawn('mmgen-tool',self.eth_args_noquiet + ['twimport',fn] + add_args)
		t.expect('(y/N): ','y')
		if expect_str:
			t.expect(expect_str)
		t.written_to_file('tracking wallet data')
		return t

	def twimport_nochksum(self):
		return self.twimport(add_args=['ignore_checksum=true'],expect_str='ignoring incorrect checksum')

	def tw_chktotal(self):
		self.spawn('',msg_only=True)
		from mmgen.tw.json import TwJSON
		fn = joinpath( self.tmpdir, TwJSON.Base(cfg,self.proto).dump_fn )
		res = json.loads(read_from_file(fn))
		cmp_or_die( res['data']['value'], vbal6, 'value in tracking wallet JSON dump' )
		return 'ok'

	async def twcompare(self):
		self.spawn('',msg_only=True)
		from mmgen.tw.ctl import TwCtl
		twctl = await TwCtl(cfg, self.proto, no_wallet_init=True)
		fn = twctl.tw_path
		fn_bak = fn.with_suffix('.bak.json')
		imsg('Comparing imported tracking wallet with original')
		data = [json.dumps(json.loads(f.read_text()), sort_keys=True) for f in (fn, fn_bak)]
		cmp_or_die(*data,'tracking wallets')
		return 'ok'

	def edit_json_twdump(self):
		self.spawn('',msg_only=True)
		from mmgen.tw.json import TwJSON
		fn = TwJSON.Base(cfg,self.proto).dump_fn
		text = json.loads(self.read_from_tmpfile(fn))
		token_addr = self.read_from_tmpfile('token_addr2').strip()
		text['data']['entries']['tokens'][token_addr][2][3] = f'edited comment [фубар] [{gr_uc}]'
		self.write_to_tmpfile( fn, json.dumps(text,indent=4) )
		return 'ok'

	def stop(self):
		self.spawn('',msg_only=True)
		if not cfg.no_daemon_stop:
			if not stop_test_daemons(self.proto.coin+'_rt'):
				return False
		set_vt100()
		return 'ok'
